from ...lib.features import FeatureVI
from .feature_v import SetupFeatureV
from ...balderglob import RuntimeObserver


class SetupFeatureVI(FeatureVI):
    feature_v = SetupFeatureV()

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, SetupFeatureVI, SetupFeatureVI.do_something, "enter `SetupFeatureVI.do_something`",
            category="feature")
        self.feature_v.called_from_outer_feature()

    def called_from_outer_feature(self):
        RuntimeObserver.add_entry(
            __file__, SetupFeatureVI, SetupFeatureVI.called_from_outer_feature,
            "enter `SetupFeatureVI.called_from_outer_feature`", category="feature")



import balder
from ...lib.features import FeatureII
from ...lib.features import ChildBConnection
from ...balderglob import RuntimeObserver


# this forces that the mapped vDevice has to have a `ChildBConnection` - all other of the conn-tree has to be removed
# for further processing
@balder.for_vdevice("VDeviceFeatureI", ChildBConnection)
class SetupFeatureII(FeatureII):

    class VDeviceFeatureI(FeatureII.VDeviceFeatureI):
        pass

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, SetupFeatureII, SetupFeatureII.do_something, "enter `SetupFeatureII.do_something`",
            category="feature")



import balder
from ..lib.features import VDeviceHelperFeature1, VDeviceHelperFeature2, VDeviceHelperFeature3
from ..lib.connections import MySimplySharedMemoryConnection, SimulatedParentConnection, SimulatedChildConnection
from ..lib.utils import SharedObj
from .features_setup import PyAddCalculate, PyAddProvideANumber


class SetupPythonAdd(balder.Setup):

    class Calculator(balder.Device):
        calc = PyAddCalculate()

    @balder.connect(Calculator, over_connection=balder.Connection.based_on(
        (SimulatedChildConnection.based_on(SimulatedParentConnection) & MySimplySharedMemoryConnection)))
    class NumberProvider1(balder.Device):
        n = PyAddProvideANumber()
        special_1 = VDeviceHelperFeature1()
        special_2 = VDeviceHelperFeature2()
        special_3 = VDeviceHelperFeature3()

    @balder.connect(Calculator, over_connection=MySimplySharedMemoryConnection)
    class NumberProvider2(balder.Device):
        n = PyAddProvideANumber()

    @balder.fixture(level="testcase")
    def cleanup_memory(self):
        SharedObj.shared_mem_list = []



from __future__ import annotations

from typing import Type, Union, List, Dict, TYPE_CHECKING

import inspect
import logging
from _balder.cnnrelations import OrConnectionRelation
from _balder.device import Device
from _balder.connection import Connection
from _balder.feature_replacement_mapping import FeatureReplacementMapping
from _balder.fixture_execution_level import FixtureExecutionLevel
from _balder.testresult import ResultState, BranchBodyResult, ResultSummary
from _balder.executor.basic_executable_executor import BasicExecutableExecutor
from _balder.executor.testcase_executor import TestcaseExecutor
from _balder.executor.unresolved_parametrized_testcase_executor import UnresolvedParametrizedTestcaseExecutor
from _balder.previous_executor_mark import PreviousExecutorMark
from _balder.routing_path import RoutingPath
from _balder.unmapped_vdevice import UnmappedVDevice
from _balder.feature_vdevice_mapping import FeatureVDeviceMapping
from _balder.controllers import DeviceController, VDeviceController, FeatureController, NormalScenarioSetupController
from _balder.exceptions import NotApplicableVariationException, UnclearAssignableFeatureConnectionError

if TYPE_CHECKING:
    from _balder.setup import Setup
    from _balder.feature import Feature
    from _balder.scenario import Scenario
    from _balder.controllers.scenario_controller import ScenarioController
    from _balder.controllers.setup_controller import SetupController
    from _balder.executor.scenario_executor import ScenarioExecutor
    from _balder.fixture_manager import FixtureManager


logger = logging.getLogger(__file__)


class VariationExecutor(BasicExecutableExecutor):
    """
    A VariationExecutor only contains :meth:`TestcaseExecutor` children.
    """
    fixture_execution_level = FixtureExecutionLevel.VARIATION

    def __init__(self, device_mapping: Dict[Type[Device], Type[Device]], parent: ScenarioExecutor):
        super().__init__()
        self._testcase_executors = []
        self._base_device_mapping = device_mapping
        self._parent_executor = parent
        self._fixture_manager = parent.fixture_manager

        # contains the active routings for the current variation
        self._routings: Dict[Connection, List[RoutingPath]] = {}
        # buffer variable to save the feature replacement after it was determined with
        # `determine_feature_replacement_and_vdevice_mappings()`
        self._feature_replacement: Union[None, Dict[Type[Device], FeatureReplacementMapping]] = None
        # buffer variable to save the feature replacement after it was determined with
        # `determine_feature_replacement_and_vdevice_mappings()`
        self._abs_setup_feature_vdevice_mappings: Union[None, Dict[Type[Device], FeatureVDeviceMapping]] = None

        # contains the absolute scenario device connections for the current variation
        self._abs_variation_scenario_device_connections: Union[List[Connection], None] = None
        # contains the absolute active variation connections (intersection between scenario based
        # `_abs_variation_scenario_device_connections` and virtual connection from active RoutingPath objects from
        # `_routings`)
        self._abs_variation_connections: Union[List[Connection], None] = None

        # contains the original active vdevice mappings for all scenario and setup devices (will be managed by
        # `update_active_vdevice_device_mappings_in_scenario_and_setup_devices()` and
        # `revert_active_vdevice_device_mappings_in_scenario_and_setup_devices()`)
        self._original_active_vdevice_mappings: Dict[Type[Device], FeatureVDeviceMapping] = {}

        # is True if the applicability check was done
        self._applicability_check_done = False
        #: this property holds the exception of type :class:`NotApplicableVariationException` if the variation can not
        #:  be applied, otherwise this property is None
        self._not_applicable_variation_exc = None

        # contains the result object for the BODY part of this branch
        self.body_result = BranchBodyResult(self)

    # ---------------------------------- STATIC METHODS ----------------------------------------------------------------

    # ---------------------------------- CLASS METHODS ----------------------------------------------------------------

    # ---------------------------------- PROPERTIES --------------------------------------------------------------------

    @property
    def base_instance(self) -> object:
        """
        returns the base class instance to which this executor instance belongs to
        """
        return None

    @property
    def parent_executor(self) -> ScenarioExecutor:
        return self._parent_executor

    @property
    def cur_scenario_class(self) -> Scenario:
        """property returns the current :class:`Scenario` for this variation"""
        return self._parent_executor.base_scenario_class

    @property
    def cur_scenario_controller(self) -> ScenarioController:
        """
        returns the current :class:`ScenarioController` for this variation
        """
        return self._parent_executor.base_scenario_controller

    @property
    def cur_setup_class(self) -> Setup:
        """property returns the current :class:`Setup` for this variation"""
        return self._parent_executor.parent_executor.base_setup_class

    @property
    def cur_setup_controller(self) -> SetupController:
        """
        returns the current :class:`SetupController` for this variation
        """
        return self._parent_executor.parent_executor.base_setup_controller

    @property
    def base_device_mapping(self) -> Dict[Type[Device], Type[Device]]:
        """
        property returns the device mapping which is a dictionary with the scenario devices as keys and their related
        setup devices as values
        """
        return self._base_device_mapping

    @property
    def all_child_executors(self) -> List[TestcaseExecutor | UnresolvedParametrizedTestcaseExecutor]:
        return self._testcase_executors

    @property
    def fixture_manager(self) -> FixtureManager:
        """returns the active fixture manager"""
        return self._fixture_manager

    @property
    def feature_replacement(self) -> Dict[Type[Device], FeatureReplacementMapping]:
        """
        this property is a dictionary with every scenario device as key and feature-replacement-mapping as value - the
        mappings hold at least information about the attribute name of the feature in scenario device, the old
        scenario-feature the instantiated feature from the scenario if it exists, otherwise this is None) and the
        new feature as second item (the feature of the related Setup-Device)
        """
        return self._feature_replacement

    @property
    def abs_setup_feature_vdevice_mappings(self) -> Dict[Type[Device], FeatureVDeviceMapping]:
        """returns the feature replacement that was determined with
        `determine_feature_replacement_and_vdevice_mappings()`"""
        return self._abs_setup_feature_vdevice_mappings

    @property
    def not_applicable_variation_exc(self) -> Union[NotApplicableVariationException, None]:
        """holds the :class:`NotApplicableVariationException` that describes why this variation in not applicable"""
        return self._not_applicable_variation_exc

    # ---------------------------------- PROTECTED METHODS -------------------------------------------------------------

    def _prepare_execution(self, show_discarded):
        print("    VARIATION ", end='')
        device_map_str = [f"{scenario_device.__qualname__}:{setup_device.__qualname__}"
                          for scenario_device, setup_device in self._base_device_mapping.items()]
        print(' | '.join(device_map_str))
        if show_discarded and not self.can_be_applied():
            print(f"      DISCARDED BECAUSE `{self.not_applicable_variation_exc.args[0]}`")
        else:
            self.determine_abs_variation_connections()
            self.update_scenario_device_feature_instances()
            self.update_active_vdevice_device_mappings_in_all_features()
            self.exchange_unmapped_vdevice_references()
            self.update_vdevice_referenced_feature_instances()
            self.set_conn_dependent_methods()
            self.resolve_and_exchange_unresolved_parametrization()

    def _body_execution(self, show_discarded):
        if show_discarded and not self.can_be_applied():
            # do nothing if this variation can not be applied (is discarded)
            return

        for cur_testcase_executor in self.get_testcase_executors():
            if (cur_testcase_executor.has_runnable_tests()
                    or cur_testcase_executor.has_skipped_tests()
                    or cur_testcase_executor.has_covered_by_tests()):
                cur_testcase_executor.execute()
            else:
                cur_testcase_executor.set_result_for_whole_branch(ResultState.NOT_RUN)

    def _cleanup_execution(self, show_discarded):
        if show_discarded and not self.can_be_applied():
            # do nothing if this variation can not be applied (is discarded)
            return

        self.restore_vdevice_feature_instances()
        self.restore_original_vdevice_references()
        self.revert_active_vdevice_device_mappings_in_all_features()
        self.revert_scenario_device_feature_instances()

    def _verify_applicability_trough_feature_implementation_matching(self):
        """
        This method validates, that the features in this variation are valid. For this the setup devices must
        implement all the required features of the scenario device. In other words, a feature child class must be
        implemented in the setup device for each feature in the scenario device.
        """
        for scenario_device, setup_device in self._base_device_mapping.items():
            scenario_device_orig_features = \
                DeviceController.get_for(scenario_device).get_original_instanced_feature_objects()
            for cur_scenario_feature_attr_name, cur_scenario_feature in scenario_device_orig_features.items():
                found_setup_feature_for_scenario_feature = False
                setup_device_instantiated_features = \
                    DeviceController.get_for(setup_device).get_all_instantiated_feature_objects()
                for _, cur_setup_feature in setup_device_instantiated_features.items():
                    if isinstance(cur_setup_feature, cur_scenario_feature.__class__):
                        found_setup_feature_for_scenario_feature = True
                if not found_setup_feature_for_scenario_feature:
                    raise NotApplicableVariationException(
                        f'no matching setup-level feature found for scenario-level feature '
                        f'`{cur_scenario_feature_attr_name} = {cur_scenario_feature.__class__.__name__}()` of device '
                        f'`{scenario_device.__qualname__}`')

    def _verify_applicability_trough_vdevice_feature_impl_matching(self) -> None:
        """
        This method checks for all vDevices that are in the setups/scenario features of this VariationExecutor, if their
        vDevices-Mappings (so the mapped setup-devices) implements all features that are defined in the vDevices
        """

        for cur_scenario_device, cur_replacement_mapping in self.feature_replacement.items():
            cur_setup_device = self.get_setup_device_for(scenario_device=cur_scenario_device)
            all_inner_setup_features = \
                DeviceController.get_for(cur_setup_device).get_all_instantiated_feature_objects()

            # now secure that all features are available in the corresponding setup device, that are defined in the
            #  mapped vDevice
            for _, cur_setup_feature_obj in all_inner_setup_features.items():
                related_scenario_feature_obj = \
                    cur_replacement_mapping.get_replaced_scenario_feature_for(cur_setup_feature_obj)

                # only check if this feature is required by the scenario
                if related_scenario_feature_obj is None:
                    # ignore this, because this feature is not used in the scenario
                    continue

                # get vDevice and device mapping
                partner_scenario_vdevice, partner_scenario_device = \
                    related_scenario_feature_obj.active_vdevice_device_mapping

                if partner_scenario_device is None:
                    # ignore because no mapping exist here
                    continue

                partner_setup_device = self.get_setup_device_for(scenario_device=partner_scenario_device)
                # get the related vDevice on setup view that is currently active
                mapped_setup_vdevices = [
                    cur_vdevice for cur_vdevice
                    in FeatureController.get_for(
                        cur_setup_feature_obj.__class__).get_abs_inner_vdevice_classes()
                    if issubclass(cur_vdevice, partner_scenario_vdevice)]
                if len(mapped_setup_vdevices) != 1:
                    # find no mapping for the vDevice -> not possible
                    # todo optimize this exception message
                    raise NotApplicableVariationException(
                        f'can not find a valid setup-level vDevice in setup feature '
                        f'`{cur_setup_feature_obj.__class__}`')
                # now check that the setup partner device has all features implemented that are required
                # features from the VDevice
                partner_setup_device_features = \
                    DeviceController.get_for(partner_setup_device).get_all_instantiated_feature_objects()
                mapped_setup_vdevices_instantiated_features = \
                    VDeviceController.get_for(mapped_setup_vdevices[0]).get_all_instantiated_feature_objects()
                for _, cur_vdevice_feature in mapped_setup_vdevices_instantiated_features.items():
                    # check that there exists a child feature in the setup device for every used feature in the
                    # vDevice class
                    if len([cur_device_feature for _, cur_device_feature in partner_setup_device_features.items()
                            if isinstance(cur_device_feature, cur_vdevice_feature.__class__)]) == 0:
                        raise NotApplicableVariationException(
                            f'can not find a child feature in mapped setup device `{partner_setup_device.__qualname__}`'
                            f' for required feature `{cur_vdevice_feature.__class__}` of vDevice '
                            f'`{mapped_setup_vdevices[0].__qualname__}`')

    def _verify_applicability_trough_all_valid_routings(self) -> None:
        """
        This method ensures that valid routings exist for every defined connection.

        The check is passed, if the method finds one or more valid routings for EVERY scenario-level
        :class:`Connection`.
        """
        if not self._routings:
            self.determine_absolute_scenario_device_connections()
            self.create_all_valid_routings()
        for scenario_cnn, cur_routings in self._routings.items():
            if len(cur_routings) == 0:
                raise NotApplicableVariationException(
                    f'can not find a valid routing on setup level for the connection `{scenario_cnn.get_tree_str()}` '
                    f'between scenario devices `{scenario_cnn.from_device}` and `{scenario_cnn.to_device}`')

    def _get_matching_setup_features_for(
            self,
            scenario_feature_obj: Feature,
            in_setup_device: Type[Device]
    ) -> List[Feature]:
        """
        Helper method that returns all matching setup features for the provided scenario feature in the provided setup
        device.
        """
        cur_setup_features = DeviceController.get_for(in_setup_device).get_all_instantiated_feature_objects()

        replacing_feature_candidates = [
            cur_setup_feature for cur_setup_feature in cur_setup_features.values()
            if isinstance(cur_setup_feature, scenario_feature_obj.__class__)
        ]
        active_scenario_vdev, mapped_scenario_dev = scenario_feature_obj.active_vdevice_device_mapping

        replacing_features = replacing_feature_candidates.copy()
        if mapped_scenario_dev is not None:
            # get the related setup device for the mapped scenario device (on scenario level)
            setup_dev_of_mapped_scenario_dev = self.get_setup_device_for(mapped_scenario_dev)

            # now check if there is a mapping on setup level too
            for cur_replacing_feature in replacing_feature_candidates:
                mapped_setup_vdev, mapped_setup_dev = cur_replacing_feature.active_vdevice_device_mapping
                if mapped_setup_vdev is not None and not issubclass(mapped_setup_vdev, active_scenario_vdev):
                    # drop this feature matching, because we have different vdevice mapped
                    replacing_features.remove(cur_replacing_feature)
                elif mapped_setup_dev is not None and mapped_setup_dev != setup_dev_of_mapped_scenario_dev:
                    # drop this feature matching, because it is not applicable here
                    replacing_features.remove(cur_replacing_feature)
        return replacing_features

    # ---------------------------------- METHODS -----------------------------------------------------------------------

    def testsummary(self) -> ResultSummary:
        if self.can_be_applied():
            return super().testsummary()
        return ResultSummary()

    def get_testcase_executors(self) -> List[TestcaseExecutor | UnresolvedParametrizedTestcaseExecutor]:
        """returns all sub testcase executors that belongs to this variation-executor"""
        return self._testcase_executors.copy()

    def add_testcase_executor(self, testcase_executor: TestcaseExecutor | UnresolvedParametrizedTestcaseExecutor):
        """
        This method adds a new TestcaseExecutor to the child element list of this object branch
        """
        if not isinstance(testcase_executor, (TestcaseExecutor, UnresolvedParametrizedTestcaseExecutor)):
            raise TypeError("the given object `testcase_executor` must be of type type `TestcaseExecutor` or "
                            "`UnresolvedParametrizedTestcaseExecutor`")
        if testcase_executor in self._testcase_executors:
            raise ValueError("the given object `testcase_executor` already exists in child list")
        self._testcase_executors.append(testcase_executor)

    def determine_feature_replacement_and_vdevice_mappings(self) -> None:
        """
        This method determines the :class:`Feature` replacement and the absolute vdevice mappings for this variation and
        its related features. It determines for every existing scenario-device-feature the related setup-device-feature
        class and sets this information to the property `_feature_replacement`. In addition to that it also determines
        the vDevice-Device mapping that will be assigned for every replaced feature later. This information will be
        saved in the property `_abs_vdevice_mappings`.

        .. note::
            If there is a feature class in the setup device, that is not used in the mapped scenario device, this
            feature will be added as **Autonomous-Feature**. This is required, because the features could be referenced
            in a lower child class than the **instantiated scenario feature**.

        .. note::
            The method automatically searches for the correct mappings. So if there exists more than one possible
            mappings for a feature, the method checks for the defined mapping on scenario level and also on setup level.
            If there are no or more than one possible mappings, the method raises a
            :class:`NotApplicableVariationError`. This error will also be thrown, if there is a vDevice mapping
            expected, while no vDevice mapping exists.

        :raises NotApplicableVariationError: will be thrown if this variation cannot be applied, because the setup-/
                                             scenario-device-features can not be resolved
        """
        feature_replacement = {
            scenario_dev: FeatureReplacementMapping() for scenario_dev in self.base_device_mapping.keys()
        }

        abs_setup_vdevice_mappings = {
            setup_dev: FeatureVDeviceMapping() for setup_dev in self.base_device_mapping.values()
        }
        for cur_scenario_device, cur_setup_device in self.base_device_mapping.items():

            for cur_attr_name, cur_scenario_feature_obj in \
                    DeviceController.get_for(cur_scenario_device).get_all_instantiated_feature_objects().items():
                active_scenario_vdevice, mapped_scenario_device = cur_scenario_feature_obj.active_vdevice_device_mapping

                cur_setup_feature_objs = self._get_matching_setup_features_for(
                    scenario_feature_obj=cur_scenario_feature_obj, in_setup_device=cur_setup_device
                )

                if len(cur_setup_feature_objs) != 1:
                    raise NotApplicableVariationException(
                        f'this variation can not be applicable because there was no setup feature implementation of '
                        f'`{cur_scenario_feature_obj.__class__.__name__}` (used by scenario device '
                        f'`{cur_scenario_device.__name__}`) in setup device `{cur_setup_device.__name__}`')
                cur_setup_feature_obj = cur_setup_feature_objs[0]

                all_abs_inner_vdevs_of_setup = \
                    FeatureController.get_for(cur_setup_feature_obj.__class__).get_abs_inner_vdevice_classes()
                used_setup_vdevice, mapped_setup_device = cur_setup_feature_obj.active_vdevice_device_mapping

                if mapped_scenario_device is None:
                    # we have exactly one matching candidate, but also no vDevice mapping
                    # check if the matching candidate has a vDevice mapping
                    if mapped_setup_device is None and len(all_abs_inner_vdevs_of_setup) > 0:
                        # there is no vDevice mapping on scenario and no vDevice mapping on setup level, but the
                        # feature defined vDevices -> NOT APPLICABLE
                        logger.warning(
                            f"missing vDevice mapping for feature "
                            f"`{cur_scenario_feature_obj.__class__.__name__}` (used in scenario device "
                            f"`{cur_scenario_device.__name__}` and in setup device `{cur_setup_device.__name__}`) - "
                            f"VARIATION CAN NOT BE APPLIED")
                        raise NotApplicableVariationException(
                            f'this variation can not be applied because there was no vDevice mapping given on '
                            f'scenario or on setup level for the feature '
                            f'`{cur_scenario_feature_obj.__class__.__name__}` (used by scenario device '
                            f'`{cur_scenario_device.__name__}`) in setup device `{cur_setup_device.__name__}`')

                if cur_attr_name not in feature_replacement[cur_scenario_device].attr_names:

                    # if there is a vDevice mapping on scenario level, but not on setup level, so update the
                    # VDevice-Device-Mapping there
                    if mapped_scenario_device is not None and mapped_setup_device is None:
                        # search the equivalent vDevice on setup level and use this one (we had not to check it,
                        # because check was already done in collector-stage)
                        setup_vdevices = [cur_vdevice for cur_vdevice in all_abs_inner_vdevs_of_setup
                                          if cur_vdevice.__name__ == active_scenario_vdevice.__name__]
                        used_setup_vdevice = setup_vdevices[0]
                        # set the mapping
                        abs_setup_vdevice_mappings[cur_setup_device].add(
                            feature=cur_setup_feature_obj,
                            mappings={
                                used_setup_vdevice: self.get_setup_device_for(mapped_scenario_device)
                            }
                        )
                    # if there is a vDevice mapping on setup level, but not on scenario level, so directly update the
                    # VDevice-Device-Mapping there
                    elif mapped_scenario_device is None and mapped_setup_device is not None:
                        abs_setup_vdevice_mappings[cur_setup_device].add(
                            feature=cur_setup_feature_obj,
                            mappings={
                                used_setup_vdevice: mapped_setup_device
                            }
                        )

                    feature_replacement[cur_scenario_device].add(attr_name=cur_attr_name,
                                                                 scenario_feature=cur_scenario_feature_obj,
                                                                 setup_feature=cur_setup_feature_obj)

            feature_replacement[cur_scenario_device].add_remaining_setup_features_as_autonomous(cur_setup_device)

        # set the result to internal properties
        self._feature_replacement = feature_replacement
        self._abs_setup_feature_vdevice_mappings = abs_setup_vdevice_mappings

    def get_setup_device_for(self, scenario_device: Type[Device]) -> Type[Device]:
        """
        This method returns the corresponding Setup-Device for the given Scenario-Device that is contained in the
        mapping of this VariationExecutor.

        :param scenario_device: the scenario device for that the mapped setup device should be returned

        :return: the mapped setup device
        """
        if scenario_device not in self.base_device_mapping.keys():
            raise ValueError(
                f"the given scenario device `{scenario_device.__name__}` is no part of the mapping of this "
                f"VariationExecutor object")
        return self.base_device_mapping[scenario_device]

    def get_scenario_device_for(self, setup_device: Type[Device]) -> Type[Device]:
        """
        This method returns the corresponding Scenario-Device for the given Setup-Device that is contained in the
        mapping of this VariationExecutor.

        :param setup_device: the setup device for that the mapped scenario device should be returned

        :return: the mapped scenario device
        """
        if setup_device not in self.base_device_mapping.values():
            raise ValueError(
                f"the given scenario device `{setup_device.__name__}` is no part of the mapping of this "
                f"VariationExecutor object")
        if list(self.base_device_mapping.values()).count(setup_device) > 1:
            raise KeyError("the requested setup device exists more than one time in `base_device_mapping`")
        return [cur_key for cur_key, cur_value in self.base_device_mapping.items() if cur_value == setup_device][0]

    def get_executor_for_testcase(self, testcase: callable) -> TestcaseExecutor | None:
        """
        This method searches for a TestcaseExecutor in the internal list for which the given testcase method is
        contained in

        :param testcase: the testcase class for which the executor should be searched for

        :return: returns the associated TestcaseExecutor or None if none could be found for the transferred type
        """
        for cur_testcase_executor in self._testcase_executors:
            if cur_testcase_executor.base_testcase_callable == testcase:
                return cur_testcase_executor
        # can not find some
        return None

    def cleanup_empty_executor_branches(self, consider_discarded=False):
        """
        This method searches the whole tree and removes branches where an executor item has no own children. It can
        remove these branches, because they have no valid matchings.

        This method implementation of the :class:`VariationExecutor` does nothing.
        """

    def update_scenario_device_feature_instances(self):
        """
        This method ensures that the (mostly abstract) feature instances of a scenario are exchanged with the
        feature instances of the assigned setup devices
        """
        for cur_scenario_device, cur_replacement_mapping in self.feature_replacement.items():
            for cur_feature_mapping in cur_replacement_mapping.mappings:
                setattr(cur_scenario_device, cur_feature_mapping.attr_name, cur_feature_mapping.setup_feature)

    def revert_scenario_device_feature_instances(self):
        """
        This method ensures that all initialized feature instances of a scenario are set back to the initial given
        features.
        """
        for cur_scenario_device, cur_replacement_mapping in self.feature_replacement.items():
            for cur_feature_mapping in cur_replacement_mapping.mappings:
                setattr(cur_scenario_device, cur_feature_mapping.attr_name, cur_feature_mapping.scenario_feature)

    def update_active_vdevice_device_mappings_in_all_features(self):
        """
        This method ensures that the correct `active_vdevices` property in all feature classes of the related setup and
        scenario-device classes are set correctly.
        """

        for cur_setup_device, feature_vdevice_mapping in self.abs_setup_feature_vdevice_mappings.items():
            if cur_setup_device not in self._original_active_vdevice_mappings.keys():
                self._original_active_vdevice_mappings[cur_setup_device] = FeatureVDeviceMapping()
            for cur_setup_feature in feature_vdevice_mapping.features:
                vdev_dev_mappings_of_setup_feat = feature_vdevice_mapping.get_mappings_for_feature(cur_setup_feature)

                cur_setup_feature_vdevice = vdev_dev_mappings_of_setup_feat[0].vdevice
                cur_mapped_setup_device = vdev_dev_mappings_of_setup_feat[0].device

                # first save old value to revert it later
                self._original_active_vdevice_mappings[cur_setup_device].add(
                    feature=cur_setup_feature,
                    mappings=cur_setup_feature.active_vdevices
                )
                # now set new value
                cur_setup_feature.active_vdevices = {cur_setup_feature_vdevice: cur_mapped_setup_device}

                # now also determine the mapping for the scenario-feature (if there exists one)
                cur_scenario_device = self.get_scenario_device_for(cur_setup_device)
                cur_scenario_feature = self.feature_replacement[cur_scenario_device].get_replaced_scenario_feature_for(
                    cur_setup_feature
                )
                if cur_scenario_feature is None:
                    # there exists no scenario feature -> we can ignore this
                    pass
                else:
                    cur_scenario_feature_controller = FeatureController.get_for(cur_scenario_feature.__class__)
                    # now get the same vdevice (same name) but on scenario level
                    cur_scenario_feature_vdevice = [
                        cur_vdevice for cur_vdevice in cur_scenario_feature_controller.get_abs_inner_vdevice_classes()
                        if cur_vdevice.__name__ == cur_setup_feature_vdevice.__name__]
                    if len(cur_scenario_feature_vdevice) == 1:
                        # only if there exists exactly one scenario vdevice with the same name

                        if cur_scenario_device not in self._original_active_vdevice_mappings.keys():
                            self._original_active_vdevice_mappings[cur_scenario_device] = FeatureVDeviceMapping()
                        # first save old value to revert it later
                        self._original_active_vdevice_mappings[cur_scenario_device].add(
                            feature=cur_scenario_feature,
                            mappings=cur_scenario_feature.active_vdevices
                        )
                        # now set new value
                        cur_scenario_feature.active_vdevices = \
                            {cur_scenario_feature_vdevice[0]: self.get_scenario_device_for(cur_mapped_setup_device)}

    def revert_active_vdevice_device_mappings_in_all_features(self):
        """
        This method ensures that the `active_vdevices` property that was changed with
        `update_active_vdevice_device_mappings_in_all_features()` will be reverted correctly.
        """
        for cur_feature_vdevice_mapping in self._original_active_vdevice_mappings.values():
            for cur_feature in cur_feature_vdevice_mapping.features:
                cur_original_mapping = {
                    mapping.vdevice:mapping.device
                    for mapping in cur_feature_vdevice_mapping.get_mappings_for_feature(cur_feature)
                }
                cur_feature.active_vdevices = cur_original_mapping

    def exchange_unmapped_vdevice_references(self):
        """
        This method exchanges all :class:`VDevice` references to an instance of :class:`UnmappedVDevice` if the
        :class:`VDevice` is not in an active VDevice-mapping.
        """
        all_devices = NormalScenarioSetupController.get_for(
            self.cur_scenario_class.__class__).get_all_abs_inner_device_classes()
        all_devices += NormalScenarioSetupController.get_for(
            self.cur_setup_class.__class__).get_all_abs_inner_device_classes()

        for cur_device in all_devices:
            cur_device_controller = DeviceController.get_for(cur_device)
            for _, cur_feature in cur_device_controller.get_all_instantiated_feature_objects().items():
                for cur_vdevice in FeatureController.get_for(cur_feature.__class__).get_abs_inner_vdevice_classes():
                    if cur_vdevice in cur_feature.active_vdevices.keys():
                        # do not exchange something, because this one is the active one
                        pass
                    else:
                        # set the `UnmappedVDevice` to this VDevice to ensure that it will not be accessed during this
                        # variation
                        setattr(cur_feature, cur_vdevice.__name__, UnmappedVDevice())

    def restore_original_vdevice_references(self):
        """
        This method restores all previously exchanged :class:`VDevice` references to the original ones.

        This method is the cleanup method for :meth:`VariationExecutor.exchange_unmapped_vdevice_references`.
        """
        all_devices = NormalScenarioSetupController.get_for(
            self.cur_scenario_class.__class__).get_all_abs_inner_device_classes()
        all_devices += NormalScenarioSetupController.get_for(
            self.cur_setup_class.__class__).get_all_abs_inner_device_classes()

        for cur_device in all_devices:
            cur_device_controller = DeviceController.get_for(cur_device)
            for _, cur_feature in cur_device_controller.get_all_instantiated_feature_objects().items():
                original_vdevices = FeatureController.get_for(cur_feature.__class__).get_original_vdevice_definitions()
                for cur_vdevice_name, cur_original_vdevice in original_vdevices.items():
                    setattr(cur_feature, cur_vdevice_name, cur_original_vdevice)

    def update_vdevice_referenced_feature_instances(self):
        """
        This method ensures that all referenced feature instances in every active vDevice classes of the used feature
        objects, will be replaced with the related feature instances of the mapped device object.

        .. note::
            Note that this method expects that the defined scenario features instances are already replaced with the
            current active setup features. In addition to that, the method expects, that the vDevice-Device mapping of
            every feature was set to the resolved setup device! This means, the method
            `update_scenario_device_feature_instances()` needs to be called before.
        """
        for scenario_device, _ in self._base_device_mapping.items():
            # these features are subclasses of the real defined one (because they are already the replaced ones)
            all_device_features = DeviceController.get_for(scenario_device).get_all_instantiated_feature_objects()
            for _, cur_feature in all_device_features.items():
                # now get the related vDevice class and update its attributes
                cur_vdevice, cur_device = cur_feature.active_vdevice_device_mapping
                if cur_vdevice is not None and cur_device is not None:
                    cur_vdevice_controller = VDeviceController.get_for(cur_vdevice)
                    cur_vdevice_all_features = cur_vdevice_controller.get_all_instantiated_feature_objects()

                    cur_device_controller = DeviceController.get_for(cur_device)
                    cur_device_all_features = cur_device_controller.get_all_instantiated_feature_objects()
                    for cur_vdevice_attr_name, cur_vdevice_feature in cur_vdevice_all_features.items():
                        # now search the used feature in the mapped device itself
                        potential_candidates = [
                            candidate_feature for _, candidate_feature in cur_device_all_features.items()
                            if isinstance(candidate_feature, cur_vdevice_feature.__class__)]
                        if len(potential_candidates) != 1:
                            raise RuntimeError("found none or more than one potential replacing candidates")
                        replacing_candidate = potential_candidates[0]
                        # because `cur_feature` is only the object instance, the value will be overwritten only for this
                        # object
                        setattr(cur_vdevice, cur_vdevice_attr_name, replacing_candidate)

    def restore_vdevice_feature_instances(self):
        """
        This method restores all previously exchanged :class:`VDevice` feature references to the original ones.

        This method is the cleanup method for :meth:`VariationExecutor.update_vdevice_referenced_feature_instances`.

        .. note::
            Note that this method expects that the defined scenario features are still replaced with the active
            setup features. In addition to that, the method expects, that the vDevice-Device mapping of every feature
            is still set!

        """
        for scenario_device, _ in self._base_device_mapping.items():
            # these features are subclasses of the real defined one (because they are already the replaced ones)
            all_device_features = DeviceController.get_for(scenario_device).get_all_instantiated_feature_objects()
            for _, cur_feature in all_device_features.items():
                # now get the related vDevice class and update its attributes
                cur_vdevice, cur_device = cur_feature.active_vdevice_device_mapping
                if cur_vdevice is not None and cur_device is not None:
                    cur_vdevice_controller = VDeviceController.get_for(cur_vdevice)
                    original_features_of_cur_vdev = cur_vdevice_controller.get_original_instanced_feature_objects()

                    for cur_vdev_attr_name, cur_vdev_orig_feature in original_features_of_cur_vdev.items():
                        setattr(cur_vdevice, cur_vdev_attr_name, cur_vdev_orig_feature)

    def verify_applicability(self) -> None:
        """
        This method verifies if this variation is executable. First the method checks if all defined
        :class:`Feature` instances are available and fully implemented in the mapped setup :class:`Device`.
        Furthermore, it checks if their exists a valid routing which also matches the defined class `@for_vdevice`
        definition of the used :class:`Feature` classes.
        """
        self._applicability_check_done = True
        try:
            self.determine_feature_replacement_and_vdevice_mappings()

            self._verify_applicability_trough_feature_implementation_matching()

            self._verify_applicability_trough_vdevice_feature_impl_matching()

            self._verify_applicability_trough_all_valid_routings()
        except NotApplicableVariationException as not_applicable_variation_exc:
            # this variation can not be used, because the features can not be resolved correctly!
            self._not_applicable_variation_exc = not_applicable_variation_exc
            self.prev_mark = PreviousExecutorMark.DISCARDED

    def can_be_applied(self) -> bool:
        """
        :return: returns True if the previous verify_applicability check was successfully
        """
        if self._applicability_check_done is not True:
            raise RuntimeError('this method can not be used before no check was executed')

        return self._not_applicable_variation_exc is None

    def determine_absolute_scenario_device_connections(self):
        """
        This method determines the absolute connections for this variation and sets the internal properties
        `_abs_variation_*_device_connections` with them. This will be used to determine the real connection-subtree
        (that can be used for this variation) by the method `create_all_valid_routings()`.

        The method re-executes the algorithm to determine the absolute connections for a scenario/setup (see the method
        :meth:`Collector.determine_absolute_device_connections_for`), but it considers the real applied vDevice and
        their feature restrictions too.
        """
        # first determine all relevant absolute connection depending on the current scenario
        abs_var_scenario_device_cnns = self.cur_scenario_controller.get_all_abs_connections()

        # now iterate over every feature, that is used by the scenario and determine the class-based feature connections
        # of the mapped scenario feature (and its vDevice)
        for cur_setup_device, feature_vdev_mapping in self.abs_setup_feature_vdevice_mappings.items():
            cur_scenario_device = self.get_scenario_device_for(cur_setup_device)
            for cur_setup_feature, vdev_mappings_of_setup_feature in feature_vdev_mapping.items():
                cur_scenario_feature: Feature = (
                    self.feature_replacement[cur_scenario_device].get_replaced_scenario_feature_for(
                        setup_feature=cur_setup_feature)
                )

                if vdev_mappings_of_setup_feature[0].device not in self.base_device_mapping.values():
                    raise NotApplicableVariationException(
                        f'the mapped setup device `{vdev_mappings_of_setup_feature[0].device.__qualname__}` which is '
                        f'mapped to the VDevice `{vdev_mappings_of_setup_feature[0].vdevice.__qualname__}` is no part '
                        f'of this variation')

                cur_mapped_scenario_device = self.get_scenario_device_for(vdev_mappings_of_setup_feature[0].device)

                # get relevant class based connections for the current feature on setup level (this is really be used
                # here)
                feat_cnn = FeatureController.get_for(cur_setup_feature.__class__)\
                    .get_abs_class_based_for_vdevice()[vdev_mappings_of_setup_feature[0].vdevice]
                # connection that are relevant for this feature
                relevant_cnns = [
                    cnn for cnn in abs_var_scenario_device_cnns
                    if (cnn.has_connection_from_to(cur_scenario_device, end_device=cur_mapped_scenario_device)
                        and max(single_feat_cnn.contained_in(cnn, ignore_metadata=True)
                                for single_feat_cnn in feat_cnn.get_singles())
                        )
                ]

                if len(relevant_cnns) > 1:
                    raise UnclearAssignableFeatureConnectionError(
                        f"the devices {cur_scenario_device.__name__} and {cur_mapped_scenario_device.__name__} have "
                        f"multiple parallel connections - the device `{cur_scenario_device.__name__}` uses a feature "
                        f"`{cur_scenario_feature.__class__.__name__}` that matches with the device "
                        f"`{cur_mapped_scenario_device.__name__}`, but it is not clear which of the parallel "
                        f"connection could be used")

                if len(relevant_cnns) == 0:
                    # todo this does not map here
                    raise ValueError("can not find matching connection on scenario level")

                relevant_device_cnn = relevant_cnns[0]

                # now cleanup the scenario-device connection `relevant_device_cnn` according to the class-based feature
                # connection
                new_cnn_to_replace = Connection.based_on(OrConnectionRelation(*[
                    cur_old_cnn_single for cur_old_cnn_single in relevant_device_cnn.get_singles()
                    if feat_cnn.contained_in(cur_old_cnn_single, ignore_metadata=True)
                ]))
                new_cnn_to_replace.set_metadata_for_all_subitems(relevant_device_cnn.metadata)

                abs_var_scenario_device_cnns.remove(relevant_device_cnn)
                abs_var_scenario_device_cnns.append(new_cnn_to_replace)

                # we do not need to check other direction because `has_connection_from_to()` returns both possibilities

        # set the determined values in variation object
        self._abs_variation_scenario_device_connections = abs_var_scenario_device_cnns

    def create_all_valid_routings(self):
        """
        This method determines all valid routings for the current variation. It iterates over every defined
        absolute determined :class:`Connection` in the :class:`Scenario` class and trys to find a valid routing for it.

        .. note::
            The method assigns a full report with all valid routings into the local property `_routings`.
        """
        self._routings = {}
        for scenario_device, _ in self._base_device_mapping.items():
            for cur_scenario_conn in self._abs_variation_scenario_device_connections:
                if cur_scenario_conn.has_connection_from_to(scenario_device):
                    # only check if the connection object has not yet been investigated
                    if cur_scenario_conn not in self._routings.keys():
                        # try to find a routing for the current connection
                        founded_routings = RoutingPath.route_through(
                            cur_scenario_conn, self._base_device_mapping)
                        self._routings[cur_scenario_conn] = founded_routings

    def determine_abs_variation_connections(self):
        """
        This method determines the absolute variation connections, that will be saved in the property
        :meth:`VariationExecutor._abs_variation_connections`. This is the INTERSECTION between the active absolute
        scenario connection for this variation (see
        :meth:`VariationExecutor._abs_variation_scenario_device_connections`) and the virtual possible connections of
        the possible routings (multiple routings will be combined over an OR relation).

        This determined connection can directly be used to determine active method variations.
        """
        virtual_routing_cnns = {}
        for cur_cnn, cur_routing_list in self._routings.items():
            virtual_routing_cnns[cur_cnn] = None
            for cur_routing in cur_routing_list:
                virtual_cnn = cur_routing.get_virtual_connection()

                if virtual_routing_cnns[cur_cnn] is None:
                    virtual_routing_cnns[cur_cnn] = Connection.based_on(virtual_cnn)
                else:
                    virtual_routing_cnns[cur_cnn] = Connection.based_on(
                        OrConnectionRelation(virtual_routing_cnns[cur_cnn], virtual_cnn))
                virtual_routing_cnns[cur_cnn].set_metadata_for_all_subitems(virtual_cnn.metadata)

        self._abs_variation_connections = []
        for cur_cnn in self._abs_variation_scenario_device_connections:
            cur_virtual_cnn = virtual_routing_cnns[cur_cnn]
            new_intersection = cur_cnn.intersection_with(cur_virtual_cnn)
            if new_intersection:
                new_intersection.set_metadata_for_all_subitems(None)
                # always set the metadata for setup devices
                new_intersection.set_metadata_for_all_subitems(cur_virtual_cnn.metadata)
            self._abs_variation_connections.append(new_intersection)

    def set_conn_dependent_methods(self):
        """
        This method sets a clear method instance for all method variations.

        .. note::
            It is important to call the method :meth:`VariationExecutor.determine_abs_variation_connections` and
            the method :meth:`VariationExecutor.update_scenario_device_feature_instances` before!
        """
        for _, setup_device in self._base_device_mapping.items():
            setup_device_instantiated_features = \
                DeviceController.get_for(setup_device).get_all_instantiated_feature_objects()
            for _, cur_setup_feature in setup_device_instantiated_features.items():

                cur_setup_feature_controller = FeatureController.get_for(cur_setup_feature.__class__)
                method_var_data_of_feature = cur_setup_feature_controller.get_method_based_for_vdevice()
                if method_var_data_of_feature is None:
                    # ignore if no method-variations exists for the current feature
                    continue

                method_var_selection = {}

                for cur_method_name, _ in inspect.getmembers(
                        cur_setup_feature, lambda o: inspect.isfunction(o) or inspect.ismethod(o)):

                    if cur_method_name not in method_var_data_of_feature.keys():
                        # ignore if there was no method-variation registration for the current method
                        continue

                    mapped_vdevice, mapped_setup_device = cur_setup_feature.active_vdevice_device_mapping
                    # get all absolute connections for this setup device
                    relevant_abs_conn = []
                    for cur_cnn in self._abs_variation_connections:
                        if cur_cnn.has_connection_from_to(start_device=setup_device, end_device=mapped_setup_device):
                            # add the children
                            relevant_abs_conn.extend(
                                cur_cnn.based_on_elements.connections if cur_cnn.__class__ == Connection else [cur_cnn]
                            )

                    if len(relevant_abs_conn) is None:
                        raise RuntimeError(f"detect empty absolute connection between device `{setup_device.__name__}` "
                                           f"and device `{mapped_setup_device.__name__}`")
                    absolute_feature_method_var_cnn = Connection.based_on(OrConnectionRelation(*relevant_abs_conn))
                    cur_method_variation = cur_setup_feature_controller.get_method_variation(
                        of_method_name=cur_method_name, for_vdevice=mapped_vdevice,
                        with_connection=absolute_feature_method_var_cnn)
                    if cur_method_variation is None:
                        raise AttributeError(f"can not find a valid method variation for method `{cur_method_name}` "
                                             f"of feature `{cur_setup_feature.__class__.__name__}`")
                    method_var_selection[cur_method_name] = \
                        (mapped_vdevice, absolute_feature_method_var_cnn, cur_method_variation)

                cur_setup_feature_controller.set_active_method_variation(method_selection=method_var_selection)

    def resolve_and_exchange_unresolved_parametrization(self):
        """resolves the parametrization if there are any :class:`UnresolvedParametrizedTestcaseExecutor` in the tree"""
        replaced_executors = []
        for cur_child in self._testcase_executors:
            if isinstance(cur_child, UnresolvedParametrizedTestcaseExecutor):
                replaced_executors.extend(cur_child.get_resolved_parametrized_testcase_executors())
            else:
                replaced_executors.append(cur_child)
        self._testcase_executors = replaced_executors



from _balder.balder_session import BalderSession
from _balder.exceptions import IllegalVDeviceMappingError, BalderException

from tests.test_utilities.base_0_envtester_class import Base0EnvtesterClass


class Test0VdeviceMappingInVdeviceFeatureScenario(Base0EnvtesterClass):
    """
    This testcase executes the basic example and checks if the tree ends with the result SUCCESS. This environment has
    one special usage of a vDevice in scenario based feature ``FeatureII``. A feature of this vDevice has an own mapping
    to another vDevice. vDevice-Mappings for features inside vDevices are not allowed! Balder should throw an error on
    collecting level.
    """

    @property
    def expected_data(self) -> tuple:
        return ()

    @property
    def expected_exit_code(self):
        return 4

    @staticmethod
    def handle_balder_exception(exc: BalderException):
        assert isinstance(exc, IllegalVDeviceMappingError), \
            f"receive a exception type that was not expected (expected `{IllegalVDeviceMappingError.__name__}`): " \
            f"{str(exc)}"
        assert exc.args[0] == 'the feature `FeatureI` you have instantiated in your vDevice `OtherVDevice1` of ' \
                              'feature `FeatureII` has a own vDevice mapping - vDevice mappings are allowed for ' \
                              'features on Devices only'

    @staticmethod
    def validate_finished_session(session: BalderSession):
        # check result states everywhere
        assert session.executor_tree is None, \
            "test session has a executor_tree object - should not be possible if error was already be detected on " \
            "constructor level (as expected)"



from multiprocessing import Process

from _balder.testresult import ResultState
from _balder.balder_session import BalderSession


def test_1_unmapped_vdevice_access(balder_working_dir):
    """
    This is a modified version of the simple calculator environment. This specialized environment has two VDevices
    :class:`ANumberProviderDevice1` and :class:`ANumberProviderDevice2` inside the ``lib.feature.AddCalculateFeature``,
    which are mapped in different constellations inside the scenarios and setups.

    There are tests and fixtures (variation and testcase level only) that secures that the
    `AccessToUnmappedVDeviceException` are thrown if the test/fixture tries to access a VDevice which is not mapped. It
    also validates that the features of the mapped VDevice can be accessed and returns the expected data.
    """
    proc = Process(target=processed, args=(balder_working_dir, ))
    proc.start()
    proc.join()
    assert proc.exitcode == 0, "the process terminates with an error"


def processed(env_dir):
    print("\n", flush=True)
    session = BalderSession(cmd_args=[], working_dir=env_dir)
    session.run()
    print("\n")

    assert session.executor_tree.executor_result == ResultState.SUCCESS, \
        "test session does not terminates with success"



from __future__ import annotations

from _balder.connection import Connection
from _balder.decorator_insert_into_tree import insert_into_tree


@insert_into_tree(parents=[])
class BluetoothConnection(Connection):
    """
    Balder Bluetooth connection (OSI LAYER 1)
    """


@insert_into_tree(parents=[])
class CanBusConnection(Connection):
    """
    Balder CAN-Bus connection (OSI LAYER 1)
    """


@insert_into_tree(parents=[])
class CoaxialCableConnection(Connection):
    """
    Balder Coaxial-Cable connection (OSI LAYER 1)
    """


@insert_into_tree(parents=[])
class DslConnection(Connection):
    """
    Balder DSL connection (OSI LAYER 1)
    """


@insert_into_tree(parents=[])
class RS232Connection(Connection):
    """
    Balder RS232 connection (OSI LAYER 1)
    """


@insert_into_tree(parents=[])
class RS422Connection(Connection):
    """
    Balder RS422 connection (OSI LAYER 1)
    """


@insert_into_tree(parents=[])
class RS485Connection(Connection):
    """
    Balder RS485 connection (OSI LAYER 1)
    """


@insert_into_tree(parents=[])
class IsdnConnection(Connection):
    """
    Balder ISDN connection (OSI LAYER 1)
    """


@insert_into_tree(parents=[])
class I2CConnection(Connection):
    """
    Balder I2C connection (OSI LAYER 1)
    """


@insert_into_tree(parents=[])
class I2SConnection(Connection):
    """
    Balder I2S connection (OSI LAYER 1)
    """


@insert_into_tree(parents=[])
class OneWireConnection(Connection):
    """
    Balder One-Wire connection (OSI LAYER 1)
    """


@insert_into_tree(parents=[])
class OpticalFiberConnection(Connection):
    """
    Balder Optical-Fiber connection (OSI LAYER 1)
    """


@insert_into_tree(parents=[])
class SpiConnection(Connection):
    """
    Balder SPI connection (OSI LAYER 1)
    """


@insert_into_tree(parents=[])
class TwistedPairCableConnection(Connection):
    """
    Balder Twisted-Pair Cable connection (OSI LAYER 1)
    """


@insert_into_tree(parents=[])
class UsbConnection(Connection):
    """
    Balder USB connection (OSI LAYER 1)
    """


@insert_into_tree(parents=[])
class WifiConnection(Connection):
    """
    Balder Wi-Fi connection (OSI LAYER 1)
    """



from multiprocessing import Process

from _balder.testresult import ResultState
from _balder.balder_session import BalderSession


def test_1_scenario_inheritance_overwrite_parent_connections(balder_working_dir):
    """
    This testcase executes the general CALCULATOR example.

    It has two inherited scenarios, while the parent scenario defines a connection and the child defines another
    connection. The environment uses a setup, that requires exactly one connection.

    The test validates, that only the single correct variation will be executed.
    This secures that the parent connection will NOT be inherited.
    """
    proc = Process(target=processed, args=(balder_working_dir, ))
    proc.start()
    proc.join()
    assert proc.exitcode == 0, "the process terminates with an error"


def processed(env_dir):
    print("\n", flush=True)
    session = BalderSession(cmd_args=[], working_dir=env_dir)
    session.run()
    print("\n")

    assert session.executor_tree.executor_result == ResultState.SUCCESS, \
        "test session does not terminates with success"
    assert len(session.executor_tree.get_setup_executors()) == 1, \
        "not exactly one setup executor found"
    assert session.executor_tree.get_setup_executors()[0].base_setup_class.__class__.__name__ == \
           "SetupPythonAdd", "wrong setup class was executed"

    scenario_executors = session.executor_tree.get_setup_executors()[0].get_scenario_executors()

    assert len(scenario_executors) == 1, \
        "not exactly one scenario executor found"
    assert scenario_executors[0].base_scenario_class.__class__.__name__ == "ScenarioAddingChild", \
        "wrong scenario class was executed"
    assert len(scenario_executors[0].get_variation_executors()) == 1, "not exactly one variation executor found"



import re
from typing import Union

from _balder.testresult import ResultState
from _balder.balder_session import BalderSession
from ...test_utilities.base_0_envtester_class import Base0EnvtesterClass
from ...test_utilities.output_compare import compare_tree_output_lines


class Test0ShowDiscarded(Base0EnvtesterClass):
    """
    This testcase executes the basic ENV example with a specific command line argument
    ``--show-discarded``. It makes sure, that the tree ends with the result SUCCESS and the additional information
    regarding the discarded variations is printed correctly.
    """

    @property
    def cmd_args(self):
        return ['--show-discarded']

    def validate_printed_output(self, stdout: str) -> bool:
        assert self._check_header_of_stdout(stdout, 2, 2, 2, expected_discarded=6), \
            f"problems within header output"

        stdout_lines = stdout.splitlines()
        assert stdout_lines[7] == "================================================== START TESTSESSION ==================================================="

        expected_output = [
            (
                'SETUP SetupB',
                [
                    (
                        '  SCENARIO ScenarioA',
                        [
                            (
                                '    VARIATION ScenarioA.ScenarioDevice1:SetupB.SetupDevice3 | ScenarioA.ScenarioDevice2:SetupB.SetupDevice4',
                                '      DISCARDED BECAUSE `this variation can not be applicable because there was no setup feature implementation of `FeatureI` (used by scenario device `ScenarioDevice1`) in setup device `SetupDevice3``'
                            ),
                            (
                                '    VARIATION ScenarioA.ScenarioDevice1:SetupB.SetupDevice4 | ScenarioA.ScenarioDevice2:SetupB.SetupDevice3',
                                '      DISCARDED BECAUSE `this variation can not be applicable because there was no setup feature implementation of `FeatureI` (used by scenario device `ScenarioDevice1`) in setup device `SetupDevice4``'
                            )
                        ]
                    ),
                    (
                        '  SCENARIO ScenarioB',
                        [
                            (
                                '    VARIATION ScenarioB.ScenarioDevice3:SetupB.SetupDevice3 | ScenarioB.ScenarioDevice4:SetupB.SetupDevice4',
                                [
                                    '      TEST ScenarioB.test_b_1 [.]',
                                    '      TEST ScenarioB.test_b_2 [.]'
                                ]

                            ),
                            (
                                '    VARIATION ScenarioB.ScenarioDevice3:SetupB.SetupDevice4 | ScenarioB.ScenarioDevice4:SetupB.SetupDevice3',
                                '      DISCARDED BECAUSE `this variation can not be applicable because there was no setup feature implementation of `FeatureIII` (used by scenario device `ScenarioDevice3`) in setup device `SetupDevice4``'
                            )
                        ]
                    )
                ]
            ),
            (
                'SETUP SetupA',
                [
                    (
                        '  SCENARIO ScenarioA',
                        [
                            (
                                '    VARIATION ScenarioA.ScenarioDevice1:SetupA.SetupDevice1 | ScenarioA.ScenarioDevice2:SetupA.SetupDevice2',
                                [
                                    '      TEST ScenarioA.test_a_1 [.]',
                                    '      TEST ScenarioA.test_a_2 [.]',
                                ]
                            ),
                            (
                                '    VARIATION ScenarioA.ScenarioDevice1:SetupA.SetupDevice2 | ScenarioA.ScenarioDevice2:SetupA.SetupDevice1',
                                '      DISCARDED BECAUSE `this variation can not be applicable because there was no setup feature implementation of `FeatureI` (used by scenario device `ScenarioDevice1`) in setup device `SetupDevice2``'
                            )
                        ]
                    ),
                    (
                        '  SCENARIO ScenarioB',
                        [
                            (
                                '    VARIATION ScenarioB.ScenarioDevice3:SetupA.SetupDevice1 | ScenarioB.ScenarioDevice4:SetupA.SetupDevice2',
                                '      DISCARDED BECAUSE `this variation can not be applicable because there was no setup feature implementation of `FeatureIII` (used by scenario device `ScenarioDevice3`) in setup device `SetupDevice1``'
                            ),
                            (
                                '    VARIATION ScenarioB.ScenarioDevice3:SetupA.SetupDevice2 | ScenarioB.ScenarioDevice4:SetupA.SetupDevice1',
                                '      DISCARDED BECAUSE `this variation can not be applicable because there was no setup feature implementation of `FeatureIII` (used by scenario device `ScenarioDevice3`) in setup device `SetupDevice2``'
                            )
                        ]
                    )
                ]
            )
        ]

        assert compare_tree_output_lines(stdout_lines[8:-2], expected_output=expected_output), \
            f"difference within current output and expected output detected"

        assert stdout_lines[-2] == "================================================== FINISH TESTSESSION =================================================="
        assert stdout_lines[-1] == "TOTAL NOT_RUN: 0 | TOTAL FAILURE: 0 | TOTAL ERROR: 0 | TOTAL SUCCESS: 4 | TOTAL SKIP: 0 | TOTAL COVERED_BY: 0"
        return True

    @property
    def expected_data(self):
        return (
            # FIXTURE-CONSTRUCTION: balderglob_fixture_session
            {"file": "balderglob.py", "meth": "balderglob_fixture_session", "part": "construction"},
            [
                (
                    # FIXTURE-CONSTRUCTION: SetupA.fixture_session
                    {"cls": "SetupA", "meth": "fixture_session", "part": "construction"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                ),
                (
                    # FIXTURE-CONSTRUCTION: SetupB.fixture_session
                    {"cls": "SetupB", "meth": "fixture_session", "part": "construction"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                )
            ],
            [
                (
                    # FIXTURE-CONSTRUCTION: ScenarioA.fixture_session
                    {"cls": "ScenarioA", "meth": "fixture_session", "part": "construction"},
                    {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                ),
                (
                    # FIXTURE-CONSTRUCTION: ScenarioB.fixture_session
                    {"cls": "ScenarioB", "meth": "fixture_session", "part": "construction"},
                    {"cls": "FeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureIV", "meth": "do_something", "category": "feature"},
                )
            ],
            [
                (
                    # FIXTURE-CONSTRUCTION: balderglob_fixture_setup
                    {"file": "balderglob.py", "meth": "balderglob_fixture_setup", "part": "construction"},
                    # FIXTURE-CONSTRUCTION: SetupA.fixture_setup
                    {"cls": "SetupA", "meth": "fixture_setup", "part": "construction"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    # FIXTURE-CONSTRUCTION: ScenarioA.fixture_setup
                    {"cls": "ScenarioA", "meth": "fixture_setup", "part": "construction"},
                    {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                    # FIXTURE-CONSTRUCTION: balderglob_fixture_scenario
                    {"file": "balderglob.py", "meth": "balderglob_fixture_scenario", "part": "construction"},
                    # FIXTURE-CONSTRUCTION: SetupA.fixture_scenario
                    {"cls": "SetupA", "meth": "fixture_scenario", "part": "construction"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    # FIXTURE-CONSTRUCTION: ScenarioA.fixture_scenario
                    {"cls": "ScenarioA", "meth": "fixture_scenario", "part": "construction"},
                    {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                    # FIXTURE-CONSTRUCTION: balderglob_fixture_variation
                    {"file": "balderglob.py", "meth": "balderglob_fixture_variation", "part": "construction"},
                    # FIXTURE-CONSTRUCTION: SetupA.fixture_variation
                    {"cls": "SetupA", "meth": "fixture_variation", "part": "construction"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    # FIXTURE-CONSTRUCTION: ScenarioA.fixture_variation
                    {"cls": "ScenarioA", "meth": "fixture_variation", "part": "construction"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    [
                        (
                            # FIXTURE-CONSTRUCTION: balderglob_fixture_testcase
                            {"file": "balderglob.py", "meth": "balderglob_fixture_testcase", "part": "construction"},
                            # FIXTURE-CONSTRUCTION: SetupA.fixture_testcase
                            {"cls": "SetupA", "meth": "fixture_testcase", "part": "construction"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            # FIXTURE-CONSTRUCTION: ScenarioA.fixture_testcase
                            {"cls": "ScenarioA", "meth": "fixture_testcase", "part": "construction"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            # TESTCASE: ScenarioA.test_a_1
                            {"cls": "ScenarioA", "meth": "test_a_1"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: ScenarioA.fixture_testcase
                            {"cls": "ScenarioA", "meth": "fixture_testcase", "part": "teardown"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: SetupA.fixture_testcase
                            {"cls": "SetupA", "meth": "fixture_testcase", "part": "teardown"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: balderglob_fixture_testcase
                            {"file": "balderglob.py", "meth": "balderglob_fixture_testcase", "part": "teardown"},
                        ),
                        (
                            # FIXTURE-CONSTRUCTION: balderglob_fixture_testcase
                            {"file": "balderglob.py", "meth": "balderglob_fixture_testcase", "part": "construction"},
                            # FIXTURE-CONSTRUCTION: SetupA.fixture_testcase
                            {"cls": "SetupA", "meth": "fixture_testcase", "part": "construction"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            # FIXTURE-CONSTRUCTION: ScenarioA.fixture_testcase
                            {"cls": "ScenarioA", "meth": "fixture_testcase", "part": "construction"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            # TESTCASE: ScenarioA.test_a_2
                            {"cls": "ScenarioA", "meth": "test_a_2"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: ScenarioA.fixture_testcase
                            {"cls": "ScenarioA", "meth": "fixture_testcase", "part": "teardown"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: SetupA.fixture_testcase
                            {"cls": "SetupA", "meth": "fixture_testcase", "part": "teardown"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: balderglob_fixture_testcase
                            {"file": "balderglob.py", "meth": "balderglob_fixture_testcase", "part": "teardown"},
                        ),
                    ],
                    # FIXTURE-TEARDOWN: ScenarioA.fixture_variation
                    {"cls": "ScenarioA", "meth": "fixture_variation", "part": "teardown"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: SetupA.fixture_variation
                    {"cls": "SetupA", "meth": "fixture_variation", "part": "teardown"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: balderglob_fixture_variation
                    {"file": "balderglob.py", "meth": "balderglob_fixture_variation", "part": "teardown"},
                    # FIXTURE-TEARDOWN: ScenarioA.fixture_scenario
                    {"cls": "ScenarioA", "meth": "fixture_scenario", "part": "teardown"},
                    {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: SetupA.fixture_scenario
                    {"cls": "SetupA", "meth": "fixture_scenario", "part": "teardown"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: balderglob_fixture_scenario
                    {"file": "balderglob.py", "meth": "balderglob_fixture_scenario", "part": "teardown"},
                    # FIXTURE-TEARDOWN: ScenarioA.fixture_setup
                    {"cls": "ScenarioA", "meth": "fixture_setup", "part": "teardown"},
                    {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: SetupA.fixture_setup
                    {"cls": "SetupA", "meth": "fixture_setup", "part": "teardown"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: balderglob_fixture_setup
                    {"file": "balderglob.py", "meth": "balderglob_fixture_setup", "part": "teardown"},
                ),
                (
                    # FIXTURE-CONSTRUCTION: balderglob_fixture_setup
                    {"file": "balderglob.py", "meth": "balderglob_fixture_setup", "part": "construction"},
                    # FIXTURE-CONSTRUCTION: SetupB.fixture_setup
                    {"cls": "SetupB", "meth": "fixture_setup", "part": "construction"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                    # FIXTURE-CONSTRUCTION: ScenarioB.fixture_setup
                    {"cls": "ScenarioB", "meth": "fixture_setup", "part": "construction"},
                    {"cls": "FeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureIV", "meth": "do_something", "category": "feature"},
                    # FIXTURE-CONSTRUCTION: balderglob_fixture_scenario
                    {"file": "balderglob.py", "meth": "balderglob_fixture_scenario", "part": "construction"},
                    # FIXTURE-CONSTRUCTION: SetupB.fixture_scenario
                    {"cls": "SetupB", "meth": "fixture_scenario", "part": "construction"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                    # FIXTURE-CONSTRUCTION: ScenarioB.fixture_scenario
                    {"cls": "ScenarioB", "meth": "fixture_scenario", "part": "construction"},
                    {"cls": "FeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureIV", "meth": "do_something", "category": "feature"},
                    # FIXTURE-CONSTRUCTION: balderglob_fixture_variation
                    {"file": "balderglob.py", "meth": "balderglob_fixture_variation", "part": "construction"},
                    # FIXTURE-CONSTRUCTION: SetupB.fixture_variation
                    {"cls": "SetupB", "meth": "fixture_variation", "part": "construction"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                    # FIXTURE-CONSTRUCTION: ScenarioB.fixture_variation
                    {"cls": "ScenarioB", "meth": "fixture_variation", "part": "construction"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                    [
                        (
                            # FIXTURE-CONSTRUCTION: balderglob_fixture_testcase
                            {"file": "balderglob.py", "meth": "balderglob_fixture_testcase", "part": "construction"},
                            # FIXTURE-CONSTRUCTION: SetupB.fixture_testcase
                            {"cls": "SetupB", "meth": "fixture_testcase", "part": "construction"},
                            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                            # FIXTURE-CONSTRUCTION: ScenarioB.fixture_testcase
                            {"cls": "ScenarioB", "meth": "fixture_testcase", "part": "construction"},
                            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                            # TESTCASE: ScenarioA.test_b_1
                            {"cls": "ScenarioB", "meth": "test_b_1"},
                            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: ScenarioB.fixture_testcase
                            {"cls": "ScenarioB", "meth": "fixture_testcase", "part": "teardown"},
                            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: SetupB.fixture_testcase
                            {"cls": "SetupB", "meth": "fixture_testcase", "part": "teardown"},
                            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: balderglob_fixture_testcase
                            {"file": "balderglob.py", "meth": "balderglob_fixture_testcase", "part": "teardown"},
                        ),
                        (
                            # FIXTURE-CONSTRUCTION: balderglob_fixture_testcase
                            {"file": "balderglob.py", "meth": "balderglob_fixture_testcase", "part": "construction"},
                            # FIXTURE-CONSTRUCTION: SetupB.fixture_testcase
                            {"cls": "SetupB", "meth": "fixture_testcase", "part": "construction"},
                            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                            # FIXTURE-CONSTRUCTION: ScenarioB.fixture_testcase
                            {"cls": "ScenarioB", "meth": "fixture_testcase", "part": "construction"},
                            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                            # TESTCASE: ScenarioA.test_b_2
                            {"cls": "ScenarioB", "meth": "test_b_2"},
                            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: ScenarioB.fixture_testcase
                            {"cls": "ScenarioB", "meth": "fixture_testcase", "part": "teardown"},
                            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: SetupB.fixture_testcase
                            {"cls": "SetupB", "meth": "fixture_testcase", "part": "teardown"},
                            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: balderglob_fixture_testcase
                            {"file": "balderglob.py", "meth": "balderglob_fixture_testcase", "part": "teardown"},
                        ),
                    ],
                    # FIXTURE-TEARDOWN: ScenarioB.fixture_variation
                    {"cls": "ScenarioB", "meth": "fixture_variation", "part": "teardown"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: SetupB.fixture_variation
                    {"cls": "SetupB", "meth": "fixture_variation", "part": "teardown"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: balderglob_fixture_variation
                    {"file": "balderglob.py", "meth": "balderglob_fixture_variation", "part": "teardown"},
                    # FIXTURE-TEARDOWN: ScenarioB.fixture_scenario
                    {"cls": "ScenarioB", "meth": "fixture_scenario", "part": "teardown"},
                    {"cls": "FeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureIV", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: SetupB.fixture_scenario
                    {"cls": "SetupB", "meth": "fixture_scenario", "part": "teardown"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: balderglob_fixture_scenario
                    {"file": "balderglob.py", "meth": "balderglob_fixture_scenario", "part": "teardown"},
                    # FIXTURE-TEARDOWN: ScenarioB.fixture_setup
                    {"cls": "ScenarioB", "meth": "fixture_setup", "part": "teardown"},
                    {"cls": "FeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureIV", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: SetupB.fixture_setup
                    {"cls": "SetupB", "meth": "fixture_setup", "part": "teardown"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: balderglob_fixture_setup
                    {"file": "balderglob.py", "meth": "balderglob_fixture_setup", "part": "teardown"},
                ),
            ],
            [
                (
                    # FIXTURE-TEARDOWN: ScenarioA.fixture_session
                    {"cls": "ScenarioA", "meth": "fixture_session", "part": "teardown"},
                    {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                ),
                (
                    # FIXTURE-TEARDOWN: ScenarioB.fixture_session
                    {"cls": "ScenarioB", "meth": "fixture_session", "part": "teardown"},
                    {"cls": "FeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureIV", "meth": "do_something", "category": "feature"},
                )
            ],
            [
                (
                    # FIXTURE-TEARDOWN: SetupA.fixture_session
                    {"cls": "SetupA", "meth": "fixture_session", "part": "teardown"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                ),
                (
                    # FIXTURE-TEARDOWN: SetupA.fixture_session
                    {"cls": "SetupB", "meth": "fixture_session", "part": "teardown"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                )
            ],
            # FIXTURE-TEARDOWN: balderglob_fixture_session
            {"file": "balderglob.py", "meth": "balderglob_fixture_session", "part": "teardown"},
        )

    @staticmethod
    def validate_finished_session(session: BalderSession):
        # check result states everywhere (have to be SUCCESS everywhere
        assert session.executor_tree.executor_result == ResultState.SUCCESS, \
            "test session does not terminates with success"

        assert session.executor_tree.construct_result.result == ResultState.SUCCESS, \
            "global executor tree construct part does not set ResultState.SUCCESS"
        assert session.executor_tree.body_result.result == ResultState.SUCCESS, \
            "global executor tree body part does not set ResultState.SUCCESS"
        assert session.executor_tree.teardown_result.result == ResultState.SUCCESS, \
            "global executor tree teardown part does not set ResultState.SUCCESS"
        for cur_setup_executor in session.executor_tree.get_setup_executors():
            assert cur_setup_executor.executor_result == ResultState.SUCCESS, \
                "the setup executor does not have result SUCCESS"

            assert cur_setup_executor.construct_result.result == ResultState.SUCCESS
            assert cur_setup_executor.body_result.result == ResultState.SUCCESS
            assert cur_setup_executor.teardown_result.result == ResultState.SUCCESS

            for cur_scenario_executor in cur_setup_executor.get_scenario_executors():
                assert cur_scenario_executor.executor_result == ResultState.SUCCESS, \
                    "the scenario executor does not have result SUCCESS"

                assert cur_scenario_executor.construct_result.result == ResultState.SUCCESS
                assert cur_scenario_executor.body_result.result == ResultState.SUCCESS
                assert cur_scenario_executor.teardown_result.result == ResultState.SUCCESS

                for cur_variation_executor in cur_scenario_executor.get_variation_executors():
                    assert cur_variation_executor.executor_result == ResultState.SUCCESS, \
                        "the variation executor does not have result SUCCESS"

                    assert cur_variation_executor.construct_result.result == ResultState.SUCCESS
                    assert cur_variation_executor.body_result.result == ResultState.SUCCESS
                    assert cur_variation_executor.teardown_result.result == ResultState.SUCCESS

                    for cur_testcase_executor in cur_variation_executor.get_testcase_executors():
                        assert cur_testcase_executor.executor_result == ResultState.SUCCESS, \
                            "the testcase executor does not have result SUCCESS"

                        assert cur_testcase_executor.construct_result.result == ResultState.SUCCESS
                        assert cur_testcase_executor.body_result.result == ResultState.SUCCESS
                        assert cur_testcase_executor.teardown_result.result == ResultState.SUCCESS



from multiprocessing import Process

from _balder.testresult import ResultState
from _balder.balder_session import BalderSession


def test_1_calculator(balder_working_dir):
    """
    This testcase executes the basic example by using the `covered_by()` decorator on scenario class. This is currently
    not supported, the test checks that an expected error message is thrown.
    """
    proc = Process(target=processed, args=(balder_working_dir, ))
    proc.start()
    proc.join()
    assert proc.exitcode == 0, "the process terminates with an error"


def processed(env_dir):
    print("\n", flush=True)
    session = BalderSession(cmd_args=[], working_dir=env_dir)
    try:
        session.run()
        print("\n")
        assert False, "test session terminates without an error"
    except NotImplementedError as exc:
        assert exc.args[0] == "The covered-by decoration of other scenario classes is not supported yet"

    assert session.executor_tree is None, "test session does not terminates before collector work was done"



from __future__ import annotations
from typing import Type
import dataclasses

from .controllers.feature_controller import FeatureController
from .device import Device
from .feature import Feature
from .vdevice import VDevice


class FeatureVDeviceMapping:
    """
    helper object that stores mappings between :class:`VDevice` and :class:`Device`
    """

    @dataclasses.dataclass
    class VDeviceDeviceMapping:
        """describes the mapping of one VDevice to one Device"""
        #: the vdevice class
        vdevice: Type[VDevice]
        #: the mapped device to this VDevice
        device: Type[Device]

    @dataclasses.dataclass
    class VDevicesMapping:
        """
        stores a single mapping
        """
        #: the feature object, this mapping belongs to
        feature: Feature
        #: the available mapping for this feature (at the moment, we only support one)
        mappings: list[FeatureVDeviceMapping.VDeviceDeviceMapping]

    def __init__(self):
        self._mappings: list[FeatureVDeviceMapping.VDevicesMapping] = []

    @property
    def mappings(self) -> list[VDevicesMapping]:
        """
        returns all existing mappings
        """
        return list(self._mappings)

    @property
    def features(self) -> list[Feature]:
        """
        returns all used attribute names
        """
        return [mapping.feature for mapping in self._mappings]

    def items(self) -> list[tuple[Feature, list[VDeviceDeviceMapping]]]:
        """
        :return: a list of Feature/vdevice mapping tuples
        """
        return [(mapping.feature, mapping.mappings) for mapping in self._mappings]

    def add(self, feature: Feature, mappings: dict[Type[VDevice], Type[Device]]):
        """
        adds a new mapping

        :param feature: the :class:`Feature` class this mapping belongs to
        :param mappings: a dictionary that describes the mapping between the :class:`VDevice` and :class:`Device`. The
                         :class:`VDevice` needs to be part of the :class:`Feature`
        """
        if feature in self.features:
            raise KeyError(f'entry for feature `{feature}` already exist - can not define it multiple times')
        feature_controller = FeatureController.get_for(feature.__class__)
        vdevice_device_mappings = []
        for cur_vdev, cur_dev in mappings.items():
            if not (isinstance(cur_vdev, type) and issubclass(cur_vdev, VDevice)):
                raise TypeError(f'the provided key objects in mappings needs to be a subclass of `{VDevice.__name__}`')
            if not (isinstance(cur_dev, type) and issubclass(cur_dev, Device)):
                raise TypeError(f'the provided value objects in mappings needs to be a subclass of `{Device.__name__}`')
            if cur_vdev not in feature_controller.get_abs_inner_vdevice_classes():
                raise ValueError(f'the provided VDevice `{cur_vdev}` is not part of the provided feature `{feature}`')
            vdevice_device_mappings.append(self.VDeviceDeviceMapping(cur_vdev, cur_dev))
        new_mapping = self.VDevicesMapping(feature=feature, mappings=vdevice_device_mappings)

        self._mappings.append(new_mapping)

    def get_mappings_for_feature(self, feature: Feature) -> list[VDeviceDeviceMapping]:
        """
        returns the list with all mappings for the provided feature
        """
        for cur_item in self._mappings:
            if cur_item.feature == feature:
                return cur_item.mappings
        raise KeyError(f'entry for feature `{feature}` does not exist')



from ...balderglob import RuntimeObserver
from .feature_i import SetupFeatureI


class SetupFeatureIOverwritten(SetupFeatureI):

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, SetupFeatureIOverwritten, SetupFeatureIOverwritten.do_something,
            "enter `SetupFeatureIOverwritten.do_something`", category="feature")



import balder
from balder.connections import DnsConnection


def test_connection_from_to():
    """
    This test checks if the method :meth:`Connection.has_connection_from_to` works correctly.
    """
    class Device1(balder.Device):
        pass

    class Device2(balder.Device):
        pass

    conn = DnsConnection(from_device=Device1, from_device_node_name='ntest1', to_device=Device2,
                         to_device_node_name='ntest2')

    assert conn.has_connection_from_to(Device1), "method does not return true"
    assert conn.has_connection_from_to(Device1, end_device=Device2), "method does not return true"
    assert conn.has_connection_from_to(Device2), "method does not return true, although this is a " \
                                                 "bidirectional connection"
    assert conn.has_connection_from_to(Device2, end_device=Device1), "method does not return true, although this is a " \
                                                                     "bidirectional connection"



from __future__ import annotations
from typing import Union, Type

import inspect
from _balder.cnnrelations import AndConnectionRelation, OrConnectionRelation
from _balder.collector import Collector
from _balder.feature import Feature
from _balder.vdevice import VDevice
from _balder.connection import Connection
from _balder.controllers import FeatureController
from _balder.exceptions import DuplicateForVDeviceError, UnknownVDeviceException


def for_vdevice(
        vdevice: Union[str, Type[VDevice]],
        with_connections: Union[
            Type[Connection], Connection, AndConnectionRelation, OrConnectionRelation
        ] = Connection(),
):
    """
    With the `@for_vdevice` you can limit the decorated object for a special allowed connection tree for every existing
    vDevice. This decorator can be used to decorate whole :class:`Feature` classes just like single methods of a
    :class:`Feature` class.

    Decorated Feature classes: This controls the allowed sub-connection tree between the mapped device of the given
                               vDevice and the device class that uses the decorated feature. If the defined sub-tree
                               does not match the sub-tree that connects the both devices with each other on the setup
                               level, the feature can not be applied.

    Decorated Feature method: Similar to the class based decoration, you can specify if a method is executable with
                              the given sub-tree. Especially, at this point you are able to define your own method
                              variations. Balder will select the chosen one depending on the matching connection
                              sub-tree.

    You can find more about this in the documentation chapter :ref:`VDevices and method-variations`.

    :param vdevice: the vDevice this decorator should describe

    :param with_connections: the assigned connection trees for this class/method (default: a universal connection)
    """
    if isinstance(with_connections, Connection):
        # do nothing
        pass
    elif isinstance(with_connections, (AndConnectionRelation, OrConnectionRelation)):
        # use container connection
        with_connections = Connection.based_on(with_connections)
    elif isinstance(with_connections, type) and issubclass(with_connections, Connection):
        # instantiate it
        with_connections = with_connections()
    else:
        raise TypeError(f"the given element ``with_connection`` needs to be from type `AndConnectionRelation`, "
                        f"`OrConnectionRelation` or `Connection` - `{type(with_connections)}` is not allowed")

    # note: if `args` is an empty list - no special sub-connection-tree bindings

    if not (isinstance(vdevice, str) or (isinstance(vdevice, type) and issubclass(vdevice, VDevice))):
        raise ValueError('the given element for `vdevice` has to be a `str` or has to be a subclass of'
                         '`VDevice`')

    class ForVDeviceDecorator:
        """decorator class for `@for_vdevice` decorator"""

        def __init__(self, func):
            nonlocal vdevice
            nonlocal with_connections

            self.func = func

            # we detect a decorated non-class object -> save it and check it later in collector
            Collector.register_possible_method_variation(func, vdevice, with_connections)

        def __new__(cls, *args, **kwargs):  # pylint: disable=unused-argument
            nonlocal vdevice

            func = args[0]

            if inspect.isclass(func):
                # it must be a class decorator
                if not issubclass(func, Feature):
                    raise TypeError(f"The decorator `@for_vdevice` may only be used for `Feature` objects. This is "
                                    f"not possible for the applied class `{func.__name__}`.")

                fn_feature_controller = FeatureController.get_for(func)

                if isinstance(vdevice, str):
                    # vDevice is a string, so we have to convert it to the correct class
                    relevant_vdevices = [cur_vdevice for cur_vdevice
                                         in fn_feature_controller.get_abs_inner_vdevice_classes()
                                         if cur_vdevice.__name__ == vdevice]

                    if len(relevant_vdevices) == 0:
                        raise ValueError(
                            f"can not find a matching inner VDevice class for the given vDevice string `{vdevice}` in "
                            f"the feature class `{func.__name__}`")

                    if len(relevant_vdevices) > 1:
                        raise RuntimeError("found more than one possible vDevices - something unexpected happened")

                    vdevice = relevant_vdevices[0]
                cls_for_vdevice = fn_feature_controller.get_class_based_for_vdevice()
                if vdevice in cls_for_vdevice.keys():
                    raise DuplicateForVDeviceError(
                        f'there already exists a decorator for the vDevice `{vdevice}` in the Feature class '
                        f'`{func.__name__}`')
                if vdevice not in fn_feature_controller.get_abs_inner_vdevice_classes():
                    raise UnknownVDeviceException(
                        f"the given vDevice `{vdevice}` is no usable vDevice in Feature class `{func.__name__}`")

                cls_for_vdevice[vdevice] = with_connections
                fn_feature_controller.set_class_based_for_vdevice(cls_for_vdevice)
                # directly return the class -> we do not want to manipulate it
                return func

            # otherwise, work will be done in `__init__`
            # return this decorator object to work with
            return super().__new__(ForVDeviceDecorator)

    return ForVDeviceDecorator



from _balder.testresult import ResultState
from _balder.balder_session import BalderSession

from tests.test_utilities.base_0_envtester_class import Base0EnvtesterClass


class Test0UnclearSetupScopedFixtureReference(Base0EnvtesterClass):
    """
    This testcase executes the basic envtester example. It forces the setup to throw a validation error for the
    unclear-setup-scoped-fixture-reference problematic.

    The problem:
        A scenario scoped fixture (defined in a scenario) with the execution level SESSION references a fixture (over
        an argument) with a name that is also used for one or more setup scoped fixtures with execution level SESSION.

        For this constellation it is unclear which referenced setup scoped fixture is the correct one.

    For this the ScenarioA references a fixture `unclear_fixture`. The SetupA implements this fixture. This should
    produce this error. Both fixtures are on SESSION level.
    """

    @property
    def expected_exit_code(self) -> int:
        return 4

    @property
    def expected_data(self) -> tuple:
        return (
            # FIXTURE-CONSTRUCTION: balderglob_fixture_session
            {"file": "balderglob.py", "meth": "balderglob_fixture_session", "part": "construction"},
            # FIXTURE-CONSTRUCTION: SetupA.fixture_session
            {"cls": "SetupA", "meth": "unclear_fixture", "part": "construction"},
            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
            # error occurs here
            # FIXTURE-TEARDOWN: SetupA.fixture_session
            {"cls": "SetupA", "meth": "unclear_fixture", "part": "teardown"},
            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
            # FIXTURE-TEARDOWN: balderglob_fixture_session
            {"file": "balderglob.py", "meth": "balderglob_fixture_session", "part": "teardown"},
        )

    @staticmethod
    def validate_finished_session(session: BalderSession):
        # check result states everywhere (have to be SUCCESS everywhere
        assert session.executor_tree.executor_result == ResultState.ERROR, \
            "test session does not terminates with an ERROR state"

        assert session.executor_tree.construct_result.result == ResultState.ERROR, \
            "global executor tree construct part does not set ResultState.SUCCESS"
        assert session.executor_tree.body_result.result == ResultState.NOT_RUN, \
            "global executor tree body part does not set ResultState.NOT_RUN"
        assert session.executor_tree.teardown_result.result == ResultState.SUCCESS, \
            "global executor tree teardown part does not set ResultState.SUCCESS"
        for cur_setup_executor in session.executor_tree.get_setup_executors():
            assert cur_setup_executor.executor_result == ResultState.NOT_RUN, \
                "the setup executor does not have result NOT_RUN"

            assert cur_setup_executor.construct_result.result == ResultState.NOT_RUN
            assert cur_setup_executor.body_result.result == ResultState.NOT_RUN
            assert cur_setup_executor.teardown_result.result == ResultState.NOT_RUN

            for cur_scenario_executor in cur_setup_executor.get_scenario_executors():
                assert cur_scenario_executor.executor_result == ResultState.NOT_RUN, \
                    "the scenario executor does not have result NOT_RUN"

                assert cur_scenario_executor.construct_result.result == ResultState.NOT_RUN
                assert cur_scenario_executor.body_result.result == ResultState.NOT_RUN
                assert cur_scenario_executor.teardown_result.result == ResultState.NOT_RUN

                for cur_variation_executor in cur_scenario_executor.get_variation_executors():
                    assert cur_variation_executor.executor_result == ResultState.NOT_RUN, \
                        "the variation executor does not have result SUCCESS"

                    assert cur_variation_executor.construct_result.result == ResultState.NOT_RUN
                    assert cur_variation_executor.body_result.result == ResultState.NOT_RUN
                    assert cur_variation_executor.teardown_result.result == ResultState.NOT_RUN

                    for cur_testcase_executor in cur_variation_executor.get_testcase_executors():
                        assert cur_testcase_executor.executor_result == ResultState.NOT_RUN, \
                            "the testcase executor does not have result SUCCESS"

                        assert cur_testcase_executor.construct_result.result == ResultState.NOT_RUN
                        assert cur_testcase_executor.body_result.result == ResultState.NOT_RUN
                        assert cur_testcase_executor.teardown_result.result == ResultState.NOT_RUN



import balder
from ..lib.features import AddCalculateFeature, ProvidesANumberFeature


class ScenarioAdding(balder.Scenario):
    IGNORE = ['do_something_but_be_not_a_test']

    class Calculator(balder.Device):
        # we could add vDevices here later too -> f.e. `NumberOne="NumberOneDevice", NumberTwo="NumberTwoDevice"`
        adds = AddCalculateFeature()

    @balder.connect(Calculator, over_connection=balder.Connection)
    class NumberOneDevice(balder.Device):
        number = ProvidesANumberFeature()

    @balder.connect(Calculator, over_connection=balder.Connection)
    class NumberTwoDevice(balder.Device):
        number = ProvidesANumberFeature()

    def do_something_but_be_not_a_test(self):
        assert False, "not good if this is thrown"

    def test_add_two_numbers(self):
        self.NumberOneDevice.number.set_number(3)
        self.NumberTwoDevice.number.set_number(4)

        self.NumberOneDevice.number.sends_the_number()
        self.NumberTwoDevice.number.sends_the_number()

        self.Calculator.adds.get_numbers()
        result = self.Calculator.adds.add_numbers()
        assert result == 7, "result is not as expected"



from _balder.routing_path import RoutingPath
import balder


class Device1(balder.Device):
    pass


class Device2(balder.Device):
    pass


class Device3(balder.Device):
    pass


def test_route_loop_detection_with_2():
    """
    This test checks that the method :meth:`RoutingPath.has_loop` detects a simple 2 device loop.
    """

    cnn10_20 = balder.Connection(
        from_device=Device1, from_device_node_name='0', to_device=Device2, to_device_node_name='0')
    cnn20_10 = balder.Connection(
        from_device=Device2, from_device_node_name='0', to_device=Device1, to_device_node_name='0')

    path = RoutingPath(cnn10_20, start_device=Device1, start_node_name='0')
    path.append_element(cnn20_10)

    assert path.has_loop() is True, "does not detect a loop, while using two similar unidirectional connections"


def test_route_loop_detection_with_2_no_loop():
    """
    This test checks that the method :meth:`RoutingPath.has_loop` does not detect a loop if no devices overlap.
    """

    cnn10_20 = balder.Connection(
        from_device=Device1, from_device_node_name='0', to_device=Device2, to_device_node_name='0')
    cnn20_30 = balder.Connection(
        from_device=Device2, from_device_node_name='0', to_device=Device3, to_device_node_name='0')

    path = RoutingPath(cnn10_20, start_device=Device1, start_node_name='0')
    path.append_element(cnn20_30)

    assert path.has_loop() is False, "does detect a loop, while no loop should be detected"


def test_route_loop_detection_with_2_with_inverse_cnn():
    """
    This test checks that the method :meth:`RoutingPath.has_loop` detects a simple 2 device loop, while one connection
    is inverse.
    """

    cnn10_20 = balder.Connection(
        from_device=Device1, from_device_node_name='0', to_device=Device2, to_device_node_name='0')
    cnn10_20_second = balder.Connection(
        from_device=Device1, from_device_node_name='0', to_device=Device2, to_device_node_name='0')

    path = RoutingPath(cnn10_20, start_device=Device1, start_node_name='0')
    path.append_element(cnn10_20_second)

    assert path.has_loop() is True, "does not detect a loop, while using two similar unidirectional connections"


def test_route_loop_detection_with_3():
    """
    This test checks that the method :meth:`RoutingPath.has_loop` detects a simple 3 device loop.
    """
    cnn10_20 = balder.Connection(
        from_device=Device1, from_device_node_name='0', to_device=Device2, to_device_node_name='0')
    cnn20_30 = balder.Connection(
        from_device=Device2, from_device_node_name='0', to_device=Device3, to_device_node_name='0')
    cnn30_10 = balder.Connection(
        from_device=Device3, from_device_node_name='0', to_device=Device1, to_device_node_name='0')

    path = RoutingPath(cnn10_20, start_device=Device1, start_node_name='0')
    path.append_element(cnn20_30)
    path.append_element(cnn30_10)

    assert path.has_loop() is True, "does not detect a loop, while using three similar bidirectional connections"


def test_route_loop_detection_with_3_with_inverse_cnn():
    """
    This test checks that the method :meth:`RoutingPath.has_loop` detects a simple 3 device loop, while the second
    connection is inverse.
    """
    cnn10_20 = balder.Connection(
        from_device=Device1, from_device_node_name='0', to_device=Device2, to_device_node_name='0')
    cnn30_20 = balder.Connection(
        from_device=Device3, from_device_node_name='0', to_device=Device2, to_device_node_name='0')
    cnn30_10 = balder.Connection(
        from_device=Device3, from_device_node_name='0', to_device=Device1, to_device_node_name='0')

    path = RoutingPath(cnn10_20, start_device=Device1, start_node_name='0')
    path.append_element(cnn30_20)
    path.append_element(cnn30_10)

    assert path.has_loop() is True, "does not detect a loop, while using three similar bidirectional connections"


def test_route_loop_detection_with_3_no_loop_because_of_different_node():
    """
    This test checks that the method :meth:`RoutingPath.has_loop` does not find a loop, if one connections goes to the
    same device but to a different node.
    """
    cnn10_20 = balder.Connection(
        from_device=Device1, from_device_node_name='0', to_device=Device2, to_device_node_name='0')
    cnn20_30 = balder.Connection(
        from_device=Device2, from_device_node_name='0', to_device=Device3, to_device_node_name='0')
    cnn30_11 = balder.Connection(
        from_device=Device3, from_device_node_name='0', to_device=Device1, to_device_node_name='1')

    path = RoutingPath(cnn10_20, start_device=Device1, start_node_name='0')
    path.append_element(cnn20_30)
    path.append_element(cnn30_11)

    assert path.has_loop() is False, "does detect a loop, while one connection goes to a previously device but "



from multiprocessing import Process

from _balder.testresult import ResultState
from _balder.balder_session import BalderSession


def test_1_calculator(balder_working_dir):
    """
    This testcase executes the example, while the scenario defines a non-existing test in its SKIP marker. It is
    expected, that the executor tree does not run and balder terminates with an excepted exception.
    """
    proc = Process(target=processed, args=(balder_working_dir, ))
    proc.start()
    proc.join()
    assert proc.exitcode == 0, "the process terminates with an error"


def processed(env_dir):
    print("\n", flush=True)
    session = BalderSession(cmd_args=[], working_dir=env_dir)
    try:
        session.run()
        print("\n")
        assert False, "test session terminates without an error"
    except ValueError as exc:
        assert exc.args[0] == ("can not find the mentioned attribute `test_that_does_not_exist` from SKIP list for "
                               "scenario `<class 'env.scenarios.scenario_adding.ScenarioAdding'>`")

    assert session.executor_tree is None, "test session does not terminates before collector work was done"



from __future__ import annotations
from typing import Any, TYPE_CHECKING

from collections import OrderedDict
from _balder.utils.functions import get_method_type
from .testcase_executor import TestcaseExecutor

if TYPE_CHECKING:
    from _balder.executor.unresolved_parametrized_testcase_executor import UnresolvedParametrizedTestcaseExecutor
    from _balder.executor.variation_executor import VariationExecutor


class ParametrizedTestcaseExecutor(TestcaseExecutor):
    """
    special testcase executor for parametrized testcases
    """
    def __init__(
            self,
            testcase: callable,
            parent: VariationExecutor,
            parametrization: OrderedDict[str, Any] = None,
            unresolved_group_obj: UnresolvedParametrizedTestcaseExecutor = None
    ) -> None:
        super().__init__(testcase=testcase, parent=parent)
        self._parametrization = parametrization
        # holds a reference to the parent unresolved object (if it has dynamic parametrized components
        self._unresolved_group_obj: UnresolvedParametrizedTestcaseExecutor | None = unresolved_group_obj

    @property
    def full_test_name_str(self) -> str:
        """
        returns the name of the test method, that should be used in output
        """
        # try to get string name for parametrization
        if self._parametrization:
            # todo what if it is not convertable to string?
            parametrization_name = ';'.join([str(e) for e in self._parametrization.values()])
            return f"{super().full_test_name_str}[{parametrization_name}]"
        return super().full_test_name_str

    def get_all_test_method_args(self):
        func_type = get_method_type(self.base_testcase_obj.__class__, self.base_testcase_callable)
        all_kwargs = self.fixture_manager.get_all_attribute_values(
            self,
            self.base_testcase_obj.__class__,
            self.base_testcase_callable,
            func_type,
            ignore_attributes=self._parametrization.keys()
        )
        if self._parametrization:
            all_kwargs.update(self._parametrization)
        return all_kwargs



from __future__ import annotations
from typing import List, Dict, Tuple, Type, Union, Callable, TYPE_CHECKING

import itertools
from _balder.fixture_manager import FixtureManager
from _balder.executor.executor_tree import ExecutorTree
from _balder.executor.setup_executor import SetupExecutor
from _balder.executor.scenario_executor import ScenarioExecutor
from _balder.executor.testcase_executor import TestcaseExecutor
from _balder.executor.variation_executor import VariationExecutor
from _balder.executor.parametrized_testcase_executor import ParametrizedTestcaseExecutor
from _balder.executor.unresolved_parametrized_testcase_executor import UnresolvedParametrizedTestcaseExecutor
from _balder.previous_executor_mark import PreviousExecutorMark
from _balder.controllers import ScenarioController, SetupController

if TYPE_CHECKING:
    from _balder.setup import Setup
    from _balder.device import Device
    from _balder.scenario import Scenario
    from _balder.connection import Connection
    from _balder.plugin_manager import PluginManager


class Solver:
    """
    The solver class is the class to map corresponding :class:`Scenario` with :class:`Setup` classes. This class
    determines all possibilities in which a scenario can be mapped to a setup constellation.
    """

    def __init__(self, setups: List[Type[Setup]], scenarios: List[Type[Scenario]], connections: List[Type[Connection]],
                 fixture_manager: Union[FixtureManager, None]):
        #: contains all available setup classes
        self._all_existing_setups = setups
        #: contains all available scenario classes
        self._all_existing_scenarios = scenarios
        #: contains all available connection classes
        self._all_existing_connections = connections
        #: contains all mappings between :meth:`Scenario`'s and :meth:`Setup`'s
        #: will contain all possibilities and will then be reduced to the really applicable set using the `filter_*()`
        #: methods
        self._mapping: List[Tuple[Type[Setup], Type[Scenario], Dict[Type[Device], Type[Device]]]] = []
        self._resolving_was_executed = False

        self._fixture_manager = fixture_manager

    # ---------------------------------- STATIC METHODS ----------------------------------------------------------------

    # ---------------------------------- CLASS METHODS -----------------------------------------------------------------

    def _set_data_for_covered_by_in_tree(self, executor_tree: ExecutorTree):
        """
        This method updates the executor tree and sets all metadata for the `@covered_by()` decorator.

        :param executor_tree: the executor tree object, that should be updated
        """
        # now determine all covered_by items
        all_scenarios = executor_tree.get_all_scenario_executors()
        all_testcases = executor_tree.get_all_testcase_executors()
        all_available_testcase_functions = [
            cur_testcase_executor.base_testcase_callable for cur_testcase_executor in all_testcases]
        all_available_scenario_classes = [
            cur_scenario_executor.base_scenario_class for cur_scenario_executor in all_scenarios]

        covered_by_mapping_of_interest = {}
        for cur_scenario in all_scenarios:
            covered_by_item = cur_scenario.get_covered_by_element()
            if covered_by_item is not None:
                covered_by_mapping_of_interest[cur_scenario] = covered_by_item

        for cur_testcase in all_testcases:
            covered_by_item = cur_testcase.get_covered_by_element()
            if covered_by_item is not None:
                covered_by_mapping_of_interest[cur_testcase] = covered_by_item

        # now go throw all covered_by items and check if the destination is contained in the tree -> set prev_mark
        for cur_elem, covered_from_items in covered_by_mapping_of_interest.items():
            all_matched_covered_from_executors = []
            for cur_covered_from_item in covered_from_items:
                if cur_covered_from_item in all_available_scenario_classes:
                    all_matched_covered_from_executors.append(
                        all_scenarios[all_available_scenario_classes.index(cur_covered_from_item)])

                elif cur_covered_from_item in all_available_testcase_functions:
                    all_matched_covered_from_executors.append(
                        all_testcases[all_available_testcase_functions.index(cur_covered_from_item)])
            if len(all_matched_covered_from_executors) > 0:
                cur_elem.prev_mark = PreviousExecutorMark.COVERED_BY
                cur_elem.covered_by_executors = all_matched_covered_from_executors

    # ---------------------------------- PROPERTIES --------------------------------------------------------------------

    @property
    def all_mappings(self) -> List[Tuple[Type[Setup], Type[Scenario], Dict[Type[Device], Type[Device]]]]:
        """
        returns all still active mappings between :meth:`Scenario`'s and :meth:`Setup`'s.
        """
        if not self._mapping or self._resolving_was_executed is False:
            raise AttributeError("please call the `resolve()` method before omitting this value")
        return self._mapping

    # ---------------------------------- PROTECTED METHODS -------------------------------------------------------------

    def _get_all_unfiltered_mappings(self) -> List[Tuple[Type[Setup], Type[Scenario]]]:
        """
        This method searches for all possible hits for the internal lists `_all_existing_setups` and
        `_all_existing_scenarios`. It returns all combinations of :meth:`Setup`'s and :meth:`Scenario`'s that could
        exist.

        :return: a list with tuple pairs that defines all possible pairs of :meth:`Setup` classes and :meth:`Scenario`
                 classes
        """
        matching_list = []
        for cur_setup in self._all_existing_setups:
            for cur_scenario in self._all_existing_scenarios:
                matching_list.append((cur_setup, cur_scenario))
        return matching_list

    # ---------------------------------- METHODS -----------------------------------------------------------------------

    def get_initial_mapping(self) -> List[Tuple[Type[Setup], Type[Scenario], Dict[Type[Device], Type[Device]]]]:
        """
        This method creates the initial amount of data for `self._mapping`. Only those elements are returned where the
        :meth:`Setup` class has more or the same amount of :meth:`Device`'s than the :meth:`Scenario` class.
        The method does not reduce the set with other criteria yet. This will be done later with filter methods.
        """
        setup_scenario_matches = self._get_all_unfiltered_mappings()

        mapping = []

        for cur_setup, cur_scenario in setup_scenario_matches:

            setup_devices = SetupController.get_for(cur_setup).get_all_abs_inner_device_classes()
            scenario_devices = ScenarioController.get_for(cur_scenario).get_all_abs_inner_device_classes()
            if len(scenario_devices) <= len(setup_devices):
                # only if there are more or as many devices in the setup as in the scenario

                # go through every possible constellation
                for cur_setup_devices in itertools.permutations(setup_devices, len(scenario_devices)):
                    # get device mapping for this constellation
                    device_mapping = {scenario_devices[idx]: cur_setup_devices[idx] for idx in range(
                        0, len(scenario_devices))}
                    mapping.append((cur_setup, cur_scenario, device_mapping))
        return mapping

    def resolve(self, plugin_manager: PluginManager) -> None:  # pylint: disable=unused-argument
        """
        This method carries out the entire resolve process and saves the end result in the object property
        `self._mapping`.
        """
        # reset mapping list
        self._mapping = []
        initial_mapping = self.get_initial_mapping()
        self._mapping = initial_mapping
        self._resolving_was_executed = True

    def get_static_parametrized_testcase_executor_for(
            self,
            variation_executor: VariationExecutor,
            testcase: Callable
    ) -> List[UnresolvedParametrizedTestcaseExecutor | ParametrizedTestcaseExecutor]:
        """
        returns a list of all testcase executors, with already resolved static parametrization -
        :class:`UnresolvedParametrizedTest` will be returned for unresolved dynamic parametrization

        :param variation_executor: the current variation executor
        :param testcase: the current testcase
        """

        scenario_controller = variation_executor.parent_executor.base_scenario_controller
        if scenario_controller.get_parametrization_for(testcase) is None:
            raise ValueError(f'can not determine parametrization for test `{testcase.__qualname__}` because no '
                             f'parametrization exist')
        static_parametrization = scenario_controller.get_parametrization_for(testcase, static=True, dynamic=False)
        executor_typ = UnresolvedParametrizedTestcaseExecutor \
            if scenario_controller.get_parametrization_for(testcase, static=False, dynamic=True) \
            else ParametrizedTestcaseExecutor

        if not static_parametrization:
            return [executor_typ(testcase, parent=variation_executor)]

        # generate product
        result = []
        for cur_product in itertools.product(*static_parametrization.values()):
            cur_product_parametrization = dict(zip(static_parametrization.keys(), cur_product))
            result.append(executor_typ(testcase, variation_executor, cur_product_parametrization))
        return result

    # pylint: disable-next=unused-argument
    def get_executor_tree(self, plugin_manager: PluginManager, add_discarded=False) -> ExecutorTree:
        """
        This method builds the ExecutorTree from the resolved data and returns it

        :param plugin_manager: the related plugin manager object
        :param add_discarded: True in case discarded elements should be added to the tree, otherwise False

        :return: the executor tree is built on the basis of the mapping data
        """

        executor_tree = ExecutorTree(self._fixture_manager)

        # create all setup executor

        for cur_setup, cur_scenario, cur_device_mapping in self._mapping:
            setup_executor = executor_tree.get_executor_for_setup(setup=cur_setup)
            if setup_executor is None:
                # setup is not available -> create new SetupExecutor
                setup_executor = SetupExecutor(cur_setup, parent=executor_tree)
                executor_tree.add_setup_executor(setup_executor)

            scenario_executor = setup_executor.get_executor_for_scenario(scenario=cur_scenario)
            if scenario_executor is None:
                # scenario is not available -> create new ScenarioExecutor
                scenario_executor = ScenarioExecutor(cur_scenario, parent=setup_executor)
                setup_executor.add_scenario_executor(scenario_executor)

            variation_executor = scenario_executor.get_executor_for_device_mapping(device_mapping=cur_device_mapping)
            if variation_executor is None:
                # variation is not available -> create new VariationExecutor
                variation_executor = VariationExecutor(device_mapping=cur_device_mapping, parent=scenario_executor)
                variation_executor.verify_applicability()

                scenario_executor.add_variation_executor(variation_executor)

                for cur_testcase in scenario_executor.base_scenario_controller.get_all_test_methods():
                    # we have a parametrization for this test case
                    if scenario_executor.base_scenario_controller.get_parametrization_for(cur_testcase):
                        testcase_executors = self.get_static_parametrized_testcase_executor_for(
                            variation_executor, cur_testcase
                        )
                        for cur_testcase_executor in testcase_executors:
                            variation_executor.add_testcase_executor(cur_testcase_executor)
                    else:
                        testcase_executor = TestcaseExecutor(cur_testcase, parent=variation_executor)
                        variation_executor.add_testcase_executor(testcase_executor)

        # now filter all elements that have no child elements
        #   -> these are items that have no valid matching, because no variation can be applied for it (there are no
        #      required :class:`Feature` matching or there exists no possible routing for the variation)
        executor_tree.cleanup_empty_executor_branches(consider_discarded=add_discarded)

        self._set_data_for_covered_by_in_tree(executor_tree=executor_tree)

        return executor_tree



from _balder.testresult import ResultState
from _balder.balder_session import BalderSession

from tests.test_utilities.base_0_envtester_class import Base0EnvtesterClass


class Test0ScenarioDeviceCorrectlyOverwriteAndAddFeatureWithoutRedefining(Base0EnvtesterClass):
    """
    This testcase executes the basic envtester environment but uses an inherited scenario `ScenarioAChild1`. Within this
    scenario we define an additional feature class that was not known in parent scenario `ScenarioAParent`. In addition
    to that, the environment overwrites an existing feature within the `ScenarioAChild1`-Device that is a child class of
    the overwritten one. It will not redefine the feature `ii = FeatureII()` of `ScenarioDevice2`. This FeatureII also
    defines a VDev relation to `Device1`. It defines two child scenarios and it will be expected that both are executed.

    The test secures that it can be executed correctly and without an error.
    """

    @property
    def expected_data(self) -> tuple:
        return (
            # FIXTURE-CONSTRUCTION: balderglob_fixture_session
            {"file": "balderglob.py", "meth": "balderglob_fixture_session", "part": "construction"},
            # FIXTURE-CONSTRUCTION: SetupA.fixture_session
            {"cls": "SetupA", "meth": "fixture_session", "part": "construction"},
            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
            {"cls": "NewlySetupFeature", "meth": "do_something", "category": "feature"},
            [
                (
                    # FIXTURE-CONSTRUCTION: ScenarioAChild1.fixture_session
                    {"cls": "ScenarioAChild1", "meth": "fixture_session", "part": "construction"},
                    {"cls": "FeatureIOverwritten", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                    {"cls": "NewlyDefinedFeature", "meth": "do_something", "category": "feature"},
                ),
                (
                    # FIXTURE-CONSTRUCTION: ScenarioAChild2.fixture_session
                    {"cls": "ScenarioAChild2", "meth": "fixture_session", "part": "construction"},
                    {"cls": "FeatureIOverwritten", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                    {"cls": "NewlyDefinedFeature", "meth": "do_something", "category": "feature"},
                )
            ],
            # FIXTURE-CONSTRUCTION: balderglob_fixture_setup
            {"file": "balderglob.py", "meth": "balderglob_fixture_setup", "part": "construction"},
            # FIXTURE-CONSTRUCTION: SetupA.fixture_setup
            {"cls": "SetupA", "meth": "fixture_setup", "part": "construction"},
            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
            {"cls": "NewlySetupFeature", "meth": "do_something", "category": "feature"},
            [
                (
                    # FIXTURE-CONSTRUCTION: ScenarioAChild1.fixture_setup
                    {"cls": "ScenarioAChild1", "meth": "fixture_setup", "part": "construction"},
                    {"cls": "FeatureIOverwritten", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                    {"cls": "NewlyDefinedFeature", "meth": "do_something", "category": "feature"},
                ),
                (
                    # FIXTURE-CONSTRUCTION: ScenarioAChild2.fixture_setup
                    {"cls": "ScenarioAChild2", "meth": "fixture_setup", "part": "construction"},
                    {"cls": "FeatureIOverwritten", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                    {"cls": "NewlyDefinedFeature", "meth": "do_something", "category": "feature"},
                )
            ],
            [
                (
                    # FIXTURE-CONSTRUCTION: balderglob_fixture_scenario
                    {"file": "balderglob.py", "meth": "balderglob_fixture_scenario", "part": "construction"},
                    # FIXTURE-CONSTRUCTION: SetupA.fixture_scenario
                    {"cls": "SetupA", "meth": "fixture_scenario", "part": "construction"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    {"cls": "NewlySetupFeature", "meth": "do_something", "category": "feature"},
                    # FIXTURE-CONSTRUCTION: ScenarioAChild1.fixture_scenario
                    {"cls": "ScenarioAChild1", "meth": "fixture_scenario", "part": "construction"},
                    {"cls": "FeatureIOverwritten", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                    {"cls": "NewlyDefinedFeature", "meth": "do_something", "category": "feature"},
                    # FIXTURE-CONSTRUCTION: balderglob_fixture_variation
                    {"file": "balderglob.py", "meth": "balderglob_fixture_variation", "part": "construction"},
                    # FIXTURE-CONSTRUCTION: SetupA.fixture_variation
                    {"cls": "SetupA", "meth": "fixture_variation", "part": "construction"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    {"cls": "NewlySetupFeature", "meth": "do_something", "category": "feature"},
                    # FIXTURE-CONSTRUCTION: ScenarioAChild1.fixture_variation
                    {"cls": "ScenarioAChild1", "meth": "fixture_variation", "part": "construction"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    {"cls": "NewlySetupFeature", "meth": "do_something", "category": "feature"},
                    [
                        (
                            # FIXTURE-CONSTRUCTION: balderglob_fixture_testcase
                            {"file": "balderglob.py", "meth": "balderglob_fixture_testcase", "part": "construction"},
                            # FIXTURE-CONSTRUCTION: SetupA.fixture_testcase
                            {"cls": "SetupA", "meth": "fixture_testcase", "part": "construction"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            {"cls": "NewlySetupFeature", "meth": "do_something", "category": "feature"},
                            # FIXTURE-CONSTRUCTION: ScenarioAChild1.fixture_testcase
                            {"cls": "ScenarioAChild1", "meth": "fixture_testcase", "part": "construction"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            {"cls": "NewlySetupFeature", "meth": "do_something", "category": "feature"},
                            # TESTCASE: ScenarioAChild1.test_a_1
                            {"cls": "ScenarioAChild1", "meth": "test_a_1"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            {"cls": "NewlySetupFeature", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: ScenarioAChild1.fixture_testcase
                            {"cls": "ScenarioAChild1", "meth": "fixture_testcase", "part": "teardown"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            {"cls": "NewlySetupFeature", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: SetupA.fixture_testcase
                            {"cls": "SetupA", "meth": "fixture_testcase", "part": "teardown"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            {"cls": "NewlySetupFeature", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: balderglob_fixture_testcase
                            {"file": "balderglob.py", "meth": "balderglob_fixture_testcase", "part": "teardown"},
                        ),
                        (
                            # FIXTURE-CONSTRUCTION: balderglob_fixture_testcase
                            {"file": "balderglob.py", "meth": "balderglob_fixture_testcase", "part": "construction"},
                            # FIXTURE-CONSTRUCTION: SetupA.fixture_testcase
                            {"cls": "SetupA", "meth": "fixture_testcase", "part": "construction"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            {"cls": "NewlySetupFeature", "meth": "do_something", "category": "feature"},
                            # FIXTURE-CONSTRUCTION: ScenarioAChild1.fixture_testcase
                            {"cls": "ScenarioAChild1", "meth": "fixture_testcase", "part": "construction"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            {"cls": "NewlySetupFeature", "meth": "do_something", "category": "feature"},
                            # TESTCASE: ScenarioAChild1.test_a_2
                            {"cls": "ScenarioAChild1", "meth": "test_a_2"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            {"cls": "NewlySetupFeature", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: ScenarioAChild1.fixture_testcase
                            {"cls": "ScenarioAChild1", "meth": "fixture_testcase", "part": "teardown"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            {"cls": "NewlySetupFeature", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: SetupA.fixture_testcase
                            {"cls": "SetupA", "meth": "fixture_testcase", "part": "teardown"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            {"cls": "NewlySetupFeature", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: balderglob_fixture_testcase
                            {"file": "balderglob.py", "meth": "balderglob_fixture_testcase", "part": "teardown"},
                        ),
                    ],
                    # FIXTURE-TEARDOWN: ScenarioAChild1.fixture_variation
                    {"cls": "ScenarioAChild1", "meth": "fixture_variation", "part": "teardown"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    {"cls": "NewlySetupFeature", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: SetupA.fixture_variation
                    {"cls": "SetupA", "meth": "fixture_variation", "part": "teardown"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    {"cls": "NewlySetupFeature", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: balderglob_fixture_variation
                    {"file": "balderglob.py", "meth": "balderglob_fixture_variation", "part": "teardown"},
                    # FIXTURE-TEARDOWN: ScenarioAChild1.fixture_scenario
                    {"cls": "ScenarioAChild1", "meth": "fixture_scenario", "part": "teardown"},
                    {"cls": "FeatureIOverwritten", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                    {"cls": "NewlyDefinedFeature", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: SetupA.fixture_scenario
                    {"cls": "SetupA", "meth": "fixture_scenario", "part": "teardown"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    {"cls": "NewlySetupFeature", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: balderglob_fixture_scenario
                    {"file": "balderglob.py", "meth": "balderglob_fixture_scenario", "part": "teardown"},
                ),
                (
                    # FIXTURE-CONSTRUCTION: balderglob_fixture_scenario
                    {"file": "balderglob.py", "meth": "balderglob_fixture_scenario", "part": "construction"},
                    # FIXTURE-CONSTRUCTION: SetupA.fixture_scenario
                    {"cls": "SetupA", "meth": "fixture_scenario", "part": "construction"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    {"cls": "NewlySetupFeature", "meth": "do_something", "category": "feature"},
                    # FIXTURE-CONSTRUCTION: ScenarioAChild2.fixture_scenario
                    {"cls": "ScenarioAChild2", "meth": "fixture_scenario", "part": "construction"},
                    {"cls": "FeatureIOverwritten", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                    {"cls": "NewlyDefinedFeature", "meth": "do_something", "category": "feature"},
                    # FIXTURE-CONSTRUCTION: balderglob_fixture_variation
                    {"file": "balderglob.py", "meth": "balderglob_fixture_variation", "part": "construction"},
                    # FIXTURE-CONSTRUCTION: SetupA.fixture_variation
                    {"cls": "SetupA", "meth": "fixture_variation", "part": "construction"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    {"cls": "NewlySetupFeature", "meth": "do_something", "category": "feature"},
                    # FIXTURE-CONSTRUCTION: ScenarioAChild2.fixture_variation
                    {"cls": "ScenarioAChild2", "meth": "fixture_variation", "part": "construction"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    {"cls": "NewlySetupFeature", "meth": "do_something", "category": "feature"},
                    [
                        (
                            # FIXTURE-CONSTRUCTION: balderglob_fixture_testcase
                            {"file": "balderglob.py", "meth": "balderglob_fixture_testcase", "part": "construction"},
                            # FIXTURE-CONSTRUCTION: SetupA.fixture_testcase
                            {"cls": "SetupA", "meth": "fixture_testcase", "part": "construction"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            {"cls": "NewlySetupFeature", "meth": "do_something", "category": "feature"},
                            # FIXTURE-CONSTRUCTION: ScenarioAChild2.fixture_testcase
                            {"cls": "ScenarioAChild2", "meth": "fixture_testcase", "part": "construction"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            {"cls": "NewlySetupFeature", "meth": "do_something", "category": "feature"},
                            # TESTCASE: ScenarioAChild2.test_a_1
                            {"cls": "ScenarioAChild2", "meth": "test_a_1"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            {"cls": "NewlySetupFeature", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: ScenarioAChild2.fixture_testcase
                            {"cls": "ScenarioAChild2", "meth": "fixture_testcase", "part": "teardown"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            {"cls": "NewlySetupFeature", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: SetupA.fixture_testcase
                            {"cls": "SetupA", "meth": "fixture_testcase", "part": "teardown"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            {"cls": "NewlySetupFeature", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: balderglob_fixture_testcase
                            {"file": "balderglob.py", "meth": "balderglob_fixture_testcase", "part": "teardown"},
                        ),
                        (
                            # FIXTURE-CONSTRUCTION: balderglob_fixture_testcase
                            {"file": "balderglob.py", "meth": "balderglob_fixture_testcase", "part": "construction"},
                            # FIXTURE-CONSTRUCTION: SetupA.fixture_testcase
                            {"cls": "SetupA", "meth": "fixture_testcase", "part": "construction"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            {"cls": "NewlySetupFeature", "meth": "do_something", "category": "feature"},
                            # FIXTURE-CONSTRUCTION: ScenarioAChild2.fixture_testcase
                            {"cls": "ScenarioAChild2", "meth": "fixture_testcase", "part": "construction"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            {"cls": "NewlySetupFeature", "meth": "do_something", "category": "feature"},
                            # TESTCASE: ScenarioAChild2.test_a_2
                            {"cls": "ScenarioAChild2", "meth": "test_a_2"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            {"cls": "NewlySetupFeature", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: ScenarioAChild2.fixture_testcase
                            {"cls": "ScenarioAChild2", "meth": "fixture_testcase", "part": "teardown"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            {"cls": "NewlySetupFeature", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: SetupA.fixture_testcase
                            {"cls": "SetupA", "meth": "fixture_testcase", "part": "teardown"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            {"cls": "NewlySetupFeature", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: balderglob_fixture_testcase
                            {"file": "balderglob.py", "meth": "balderglob_fixture_testcase", "part": "teardown"},
                        ),
                    ],
                    # FIXTURE-TEARDOWN: ScenarioAChild2.fixture_variation
                    {"cls": "ScenarioAChild2", "meth": "fixture_variation", "part": "teardown"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    {"cls": "NewlySetupFeature", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: SetupA.fixture_variation
                    {"cls": "SetupA", "meth": "fixture_variation", "part": "teardown"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    {"cls": "NewlySetupFeature", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: balderglob_fixture_variation
                    {"file": "balderglob.py", "meth": "balderglob_fixture_variation", "part": "teardown"},
                    # FIXTURE-TEARDOWN: ScenarioAChild2.fixture_scenario
                    {"cls": "ScenarioAChild2", "meth": "fixture_scenario", "part": "teardown"},
                    {"cls": "FeatureIOverwritten", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                    {"cls": "NewlyDefinedFeature", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: SetupA.fixture_scenario
                    {"cls": "SetupA", "meth": "fixture_scenario", "part": "teardown"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    {"cls": "NewlySetupFeature", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: balderglob_fixture_scenario
                    {"file": "balderglob.py", "meth": "balderglob_fixture_scenario", "part": "teardown"},
                )
            ],
            [
                (
                    # FIXTURE-TEARDOWN: ScenarioAChild1.fixture_setup
                    {"cls": "ScenarioAChild1", "meth": "fixture_setup", "part": "teardown"},
                    {"cls": "FeatureIOverwritten", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                    {"cls": "NewlyDefinedFeature", "meth": "do_something", "category": "feature"},
                ),
                (
                    # FIXTURE-TEARDOWN: ScenarioAChild2.fixture_setup
                    {"cls": "ScenarioAChild2", "meth": "fixture_setup", "part": "teardown"},
                    {"cls": "FeatureIOverwritten", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                    {"cls": "NewlyDefinedFeature", "meth": "do_something", "category": "feature"},
                )
            ],
            # FIXTURE-TEARDOWN: SetupA.fixture_setup
            {"cls": "SetupA", "meth": "fixture_setup", "part": "teardown"},
            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
            {"cls": "NewlySetupFeature", "meth": "do_something", "category": "feature"},
            # FIXTURE-TEARDOWN: balderglob_fixture_setup
            {"file": "balderglob.py", "meth": "balderglob_fixture_setup", "part": "teardown"},
            [
                (
                    # FIXTURE-TEARDOWN: ScenarioAChild1.fixture_session
                    {"cls": "ScenarioAChild1", "meth": "fixture_session", "part": "teardown"},
                    {"cls": "FeatureIOverwritten", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                    {"cls": "NewlyDefinedFeature", "meth": "do_something", "category": "feature"},
                ),
                (
                    # FIXTURE-TEARDOWN: ScenarioAChild2.fixture_session
                    {"cls": "ScenarioAChild2", "meth": "fixture_session", "part": "teardown"},
                    {"cls": "FeatureIOverwritten", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                    {"cls": "NewlyDefinedFeature", "meth": "do_something", "category": "feature"},
                )
            ],
            # FIXTURE-TEARDOWN: SetupA.fixture_session
            {"cls": "SetupA", "meth": "fixture_session", "part": "teardown"},
            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
            {"cls": "NewlySetupFeature", "meth": "do_something", "category": "feature"},
            # FIXTURE-TEARDOWN: balderglob_fixture_session
            {"file": "balderglob.py", "meth": "balderglob_fixture_session", "part": "teardown"},
        )

    @staticmethod
    def validate_finished_session(session: BalderSession):

        # check result states everywhere (have to be SUCCESS everywhere
        assert session.executor_tree.executor_result == ResultState.SUCCESS, \
            "test session does not terminates with success"

        assert session.executor_tree.construct_result.result == ResultState.SUCCESS, \
            "global executor tree construct part does not set ResultState.SUCCESS"
        assert session.executor_tree.body_result.result == ResultState.SUCCESS, \
            "global executor tree body part does not set ResultState.SUCCESS"
        assert session.executor_tree.teardown_result.result == ResultState.SUCCESS, \
            "global executor tree teardown part does not set ResultState.SUCCESS"
        for cur_setup_executor in session.executor_tree.get_setup_executors():
            assert cur_setup_executor.executor_result == ResultState.SUCCESS, \
                "the setup executor does not have result SUCCESS"

            assert cur_setup_executor.construct_result.result == ResultState.SUCCESS
            assert cur_setup_executor.body_result.result == ResultState.SUCCESS
            assert cur_setup_executor.teardown_result.result == ResultState.SUCCESS

            for cur_scenario_executor in cur_setup_executor.get_scenario_executors():
                assert cur_scenario_executor.executor_result == ResultState.SUCCESS, \
                    "the scenario executor does not have result SUCCESS"

                assert cur_scenario_executor.construct_result.result == ResultState.SUCCESS
                assert cur_scenario_executor.body_result.result == ResultState.SUCCESS
                assert cur_scenario_executor.teardown_result.result == ResultState.SUCCESS

                for cur_variation_executor in cur_scenario_executor.get_variation_executors():
                    assert cur_variation_executor.executor_result == ResultState.SUCCESS, \
                        "the variation executor does not have result SUCCESS"

                    assert cur_variation_executor.construct_result.result == ResultState.SUCCESS
                    assert cur_variation_executor.body_result.result == ResultState.SUCCESS
                    assert cur_variation_executor.teardown_result.result == ResultState.SUCCESS

                    for cur_testcase_executor in cur_variation_executor.get_testcase_executors():
                        assert cur_testcase_executor.executor_result == ResultState.SUCCESS, \
                            "the testcase executor does not have result SUCCESS"

                        assert cur_testcase_executor.construct_result.result == ResultState.SUCCESS
                        assert cur_testcase_executor.body_result.result == ResultState.SUCCESS
                        assert cur_testcase_executor.teardown_result.result == ResultState.SUCCESS
